from __future__ import annotations

from collections.abc import Iterable
from inspect import isawaitable
from typing import TYPE_CHECKING, Any, Callable


if TYPE_CHECKING:
    from sanic import Sanic


def trigger_events(
    events: Iterable[Callable[..., Any]] | None,
    loop,
    app: Sanic | None = None,
    **kwargs,
):
    """Trigger event callbacks (functions or async)

    Args:
        events (Optional[Iterable[Callable[..., Any]]]): [description]
        loop ([type]): [description]
        app (Optional[Sanic], optional): [description]. Defaults to None.
    """
    if events:
        for event in events:
            try:
                result = event(**kwargs) if not app else event(app, **kwargs)
            except TypeError:
                result = (
                    event(loop, **kwargs)
                    if not app
                    else event(app, loop, **kwargs)
                )
            if isawaitable(result):
                loop.run_until_complete(result)



from .convenience import (
    empty,
    file,
    file_stream,
    html,
    json,
    raw,
    redirect,
    text,
    validate_file,
)
from .types import (
    BaseHTTPResponse,
    HTTPResponse,
    JSONResponse,
    ResponseStream,
    json_dumps,
)


__all__ = (
    "BaseHTTPResponse",
    "HTTPResponse",
    "JSONResponse",
    "ResponseStream",
    "empty",
    "json",
    "text",
    "raw",
    "html",
    "validate_file",
    "file",
    "redirect",
    "file_stream",
    "json_dumps",
)



from .form import File, parse_multipart_form
from .parameters import RequestParameters
from .types import Request


__all__ = (
    "File",
    "parse_multipart_form",
    "Request",
    "RequestParameters",
)



from __future__ import annotations

import os
import sys

from asyncio import new_event_loop
from itertools import chain
from multiprocessing.connection import Connection
from pathlib import Path
from signal import SIGINT, SIGTERM
from signal import signal as signal_func
from time import sleep

from sanic.server.events import trigger_events
from sanic.worker.loader import AppLoader


class Reloader:
    INTERVAL = 1.0  # seconds

    def __init__(
        self,
        publisher: Connection,
        interval: float,
        reload_dirs: set[Path],
        app_loader: AppLoader,
    ):
        self._publisher = publisher
        self.interval = interval or self.INTERVAL
        self.reload_dirs = reload_dirs
        self.run = True
        self.app_loader = app_loader

    def __call__(self) -> None:
        app = self.app_loader.load()
        signal_func(SIGINT, self.stop)
        signal_func(SIGTERM, self.stop)
        mtimes: dict[str, float] = {}

        reloader_start = app.listeners.get("reload_process_start")
        reloader_stop = app.listeners.get("reload_process_stop")
        before_trigger = app.listeners.get("before_reload_trigger")
        after_trigger = app.listeners.get("after_reload_trigger")
        loop = new_event_loop()
        if reloader_start:
            trigger_events(reloader_start, loop, app)

        while self.run:
            changed = set()
            for filename in self.files():
                try:
                    if self.check_file(filename, mtimes):
                        path = (
                            filename
                            if isinstance(filename, str)
                            else filename.resolve()
                        )
                        changed.add(str(path))
                except OSError:
                    continue
            if changed:
                if before_trigger:
                    trigger_events(before_trigger, loop, app)
                self.reload(",".join(changed) if changed else "unknown")
                if after_trigger:
                    trigger_events(after_trigger, loop, app, changed=changed)
            sleep(self.interval)
        else:
            if reloader_stop:
                trigger_events(reloader_stop, loop, app)

    def stop(self, *_):
        self.run = False

    def reload(self, reloaded_files):
        message = f"__ALL_PROCESSES__:{reloaded_files}"
        self._publisher.send(message)

    def files(self):
        return chain(
            self.python_files(),
            *(d.glob("**/*") for d in self.reload_dirs),
        )

    def python_files(self):  # no cov
        """This iterates over all relevant Python files.

        It goes through all
        loaded files from modules, all files in folders of already loaded
        modules as well as all files reachable through a package.
        """
        # The list call is necessary on Python 3 in case the module
        # dictionary modifies during iteration.
        for module in list(sys.modules.values()):
            if module is None:
                continue
            filename = getattr(module, "__file__", None)
            if filename:
                old = None
                while not os.path.isfile(filename):
                    old = filename
                    filename = os.path.dirname(filename)
                    if filename == old:
                        break
                else:
                    if filename[-4:] in (".pyc", ".pyo"):
                        filename = filename[:-1]
                    yield filename

    @staticmethod
    def check_file(filename, mtimes) -> bool:
        need_reload = False

        mtime = os.stat(filename).st_mtime
        old_time = mtimes.get(filename)
        if old_time is None:
            mtimes[filename] = mtime
        elif mtime > old_time:
            mtimes[filename] = mtime
            need_reload = True

        return need_reload



from __future__ import annotations

import os

from collections.abc import MutableMapping
from datetime import datetime, timezone
from inspect import signature
from multiprocessing.context import BaseContext
from signal import SIGINT
from threading import Thread
from time import sleep
from typing import Any

from sanic.log import Colors, logger
from sanic.worker.constants import ProcessState, RestartOrder


def get_now():
    now = datetime.now(tz=timezone.utc)
    return now


class WorkerProcess:
    """A worker process."""

    THRESHOLD = 300  # == 30 seconds
    SERVER_LABEL = "Server"
    SERVER_IDENTIFIER = "Srv"

    def __init__(
        self,
        factory,
        name,
        ident,
        target,
        kwargs,
        worker_state,
        restartable: bool = False,
    ):
        self.state = ProcessState.IDLE
        self.factory = factory
        self.name = name
        self.ident = ident
        self.target = target
        self.kwargs = kwargs
        self.worker_state = worker_state
        self.restartable = restartable
        if self.name not in self.worker_state:
            self.worker_state[self.name] = {
                "server": self.SERVER_LABEL in self.name
            }
        self.spawn()

    def set_state(self, state: ProcessState, force=False):
        if not force and state < self.state:
            raise Exception("...")
        self.state = state
        self.worker_state[self.name] = {
            **self.worker_state[self.name],
            "state": self.state.name,
        }

    def start(self):
        os.environ["SANIC_WORKER_NAME"] = self.name
        os.environ["SANIC_WORKER_IDENTIFIER"] = self.ident
        logger.debug(
            f"{Colors.BLUE}Starting a process: {Colors.BOLD}"
            f"{Colors.SANIC}%s{Colors.END}",
            self.name,
        )
        self.set_state(ProcessState.STARTING)
        self._current_process.start()
        self.set_state(ProcessState.STARTED)
        if not self.worker_state[self.name].get("starts"):
            self.worker_state[self.name] = {
                **self.worker_state[self.name],
                "pid": self.pid,
                "start_at": get_now(),
                "starts": 1,
            }
        del os.environ["SANIC_WORKER_NAME"]
        del os.environ["SANIC_WORKER_IDENTIFIER"]

    def join(self):
        self.set_state(ProcessState.JOINED)
        self._current_process.join()

    def exit(self):
        limit = 100
        while self.is_alive() and limit > 0:
            sleep(0.1)
            limit -= 1

        if not self.is_alive():
            try:
                del self.worker_state[self.name]
            except (
                BrokenPipeError,
                ConnectionRefusedError,
                ConnectionResetError,
                EOFError,
            ):
                logger.debug("Monitor process has already exited.")
            except KeyError:
                logger.debug("Could not find worker state to delete.")

    def terminate(self):
        if self.state is not ProcessState.TERMINATED:
            logger.debug(
                f"{Colors.BLUE}Terminating a process: "
                f"{Colors.BOLD}{Colors.SANIC}"
                f"%s {Colors.BLUE}[%s]{Colors.END}",
                self.name,
                self.pid,
            )
            try:
                self.set_state(ProcessState.TERMINATED, force=True)
            except (BrokenPipeError, ConnectionResetError, EOFError):
                pass
            try:
                os.kill(self.pid, SIGINT)
            except (KeyError, AttributeError, ProcessLookupError):
                ...

    def restart(self, restart_order=RestartOrder.SHUTDOWN_FIRST, **kwargs):
        logger.debug(
            f"{Colors.BLUE}Restarting a process: {Colors.BOLD}{Colors.SANIC}"
            f"%s {Colors.BLUE}[%s]{Colors.END}",
            self.name,
            self.pid,
        )
        self.set_state(ProcessState.RESTARTING, force=True)
        if restart_order is RestartOrder.SHUTDOWN_FIRST:
            self._terminate_now()
        else:
            self._old_process = self._current_process
        if self._add_config():
            self.kwargs.update(
                {"config": {k.upper(): v for k, v in kwargs.items()}}
            )
        try:
            self.spawn()
            self.start()
        except AttributeError:
            raise RuntimeError("Restart failed")

        if restart_order is RestartOrder.STARTUP_FIRST:
            self._terminate_soon()

        self.worker_state[self.name] = {
            **self.worker_state[self.name],
            "pid": self.pid,
            "starts": self.worker_state[self.name]["starts"] + 1,
            "restart_at": get_now(),
        }

    def is_alive(self):
        try:
            return self._current_process.is_alive()
        except AssertionError:
            return False

    def spawn(self):
        if self.state not in (ProcessState.IDLE, ProcessState.RESTARTING):
            raise Exception("Cannot spawn a worker process until it is idle.")
        self._current_process = self.factory(
            name=self.name,
            target=self.target,
            kwargs=self.kwargs,
            daemon=True,
        )

    @property
    def pid(self):
        return self._current_process.pid

    @property
    def exitcode(self):
        return self._current_process.exitcode

    def _terminate_now(self):
        if not self._current_process.is_alive():
            return
        logger.debug(
            f"{Colors.BLUE}Begin restart termination: "
            f"{Colors.BOLD}{Colors.SANIC}"
            f"%s {Colors.BLUE}[%s]{Colors.END}",
            self.name,
            self._current_process.pid,
        )
        self._current_process.terminate()

    def _terminate_soon(self):
        logger.debug(
            f"{Colors.BLUE}Begin restart termination: "
            f"{Colors.BOLD}{Colors.SANIC}"
            f"%s {Colors.BLUE}[%s]{Colors.END}",
            self.name,
            self._current_process.pid,
        )
        termination_thread = Thread(target=self._wait_to_terminate)
        termination_thread.start()

    def _wait_to_terminate(self):
        logger.debug(
            f"{Colors.BLUE}Waiting for process to be acked: "
            f"{Colors.BOLD}{Colors.SANIC}"
            f"%s {Colors.BLUE}[%s]{Colors.END}",
            self.name,
            self._old_process.pid,
        )
        misses = 0
        while self.state is not ProcessState.ACKED:
            sleep(0.1)
            misses += 1
            if misses > self.THRESHOLD:
                raise TimeoutError(
                    f"Worker {self.name} failed to come ack within "
                    f"{self.THRESHOLD / 10} seconds"
                )
        else:
            logger.debug(
                f"{Colors.BLUE}Process acked. Terminating: "
                f"{Colors.BOLD}{Colors.SANIC}"
                f"%s {Colors.BLUE}[%s]{Colors.END}",
                self.name,
                self._old_process.pid,
            )
            self._old_process.terminate()
        delattr(self, "_old_process")

    def _add_config(self) -> bool:
        sig = signature(self.target)
        if "config" in sig.parameters or any(
            param.kind == param.VAR_KEYWORD
            for param in sig.parameters.values()
        ):
            return True
        return False


class Worker:
    WORKER_PREFIX = "Sanic"

    def __init__(
        self,
        ident: str,
        name: str,
        serve,
        server_settings,
        context: BaseContext,
        worker_state: MutableMapping[str, Any],
        num: int = 1,
        restartable: bool = False,
        tracked: bool = True,
        auto_start: bool = True,
    ):
        self.ident = ident
        self.name = name
        self.num = num
        self.context = context
        self.serve = serve
        self.server_settings = server_settings
        self.worker_state = worker_state
        self.processes: set[WorkerProcess] = set()
        self.restartable = restartable
        self.tracked = tracked
        self.auto_start = auto_start
        for _ in range(num):
            self.create_process()

    def create_process(self) -> WorkerProcess:
        process = WorkerProcess(
            # Need to ignore this typing error - The problem is the
            # BaseContext itself has no Process. But, all of its
            # implementations do. We can safely ignore as it is a typing
            # issue in the standard lib.
            factory=self.context.Process,  # type: ignore
            name="-".join(
                [self.WORKER_PREFIX, self.name, str(len(self.processes))]
            ),
            ident=self.ident,
            target=self.serve,
            kwargs={**self.server_settings},
            worker_state=self.worker_state,
            restartable=self.restartable,
        )
        self.processes.add(process)
        return process

    def has_alive_processes(self) -> bool:
        return any(process.is_alive() for process in self.processes)



from __future__ import annotations

from sanic.errorpages import BaseRenderer, TextRenderer, exception_response
from sanic.exceptions import ServerError
from sanic.log import error_logger
from sanic.models.handler_types import RouteHandler
from sanic.request.types import Request
from sanic.response import text
from sanic.response.types import HTTPResponse


class ErrorHandler:
    """Process and handle all uncaught exceptions.

    This error handling framework is built into the core that can be extended
    by the developers to perform a wide range of tasks from recording the error
    stats to reporting them to an external service that can be used for
    realtime alerting system.

    Args:
        base (BaseRenderer): The renderer to use for the error pages.
    """  # noqa: E501

    def __init__(
        self,
        base: type[BaseRenderer] = TextRenderer,
    ):
        self.cached_handlers: dict[
            tuple[type[BaseException], str | None], RouteHandler | None
        ] = {}
        self.debug = False
        self.base = base

    def _full_lookup(self, exception, route_name: str | None = None):
        return self.lookup(exception, route_name)

    def _add(
        self,
        key: tuple[type[BaseException], str | None],
        handler: RouteHandler,
    ) -> None:
        if key in self.cached_handlers:
            exc, name = key
            if name is None:
                name = "__ALL_ROUTES__"

            message = (
                f"Duplicate exception handler definition on: route={name} "
                f"and exception={exc}"
            )
            raise ServerError(message)
        self.cached_handlers[key] = handler

    def add(self, exception, handler, route_names: list[str] | None = None):
        """Add a new exception handler to an already existing handler object.

        Args:
            exception (sanic.exceptions.SanicException or Exception): Type
                of exception that needs to be handled.
            handler (function): Reference to the function that will
                handle the exception.

        Returns:
            None

        """  # noqa: E501
        if route_names:
            for route in route_names:
                self._add((exception, route), handler)
        else:
            self._add((exception, None), handler)

    def lookup(self, exception, route_name: str | None = None):
        """Lookup the existing instance of `ErrorHandler` and fetch the registered handler for a specific type of exception.

        This method leverages a dict lookup to speedup the retrieval process.

        Args:
            exception (sanic.exceptions.SanicException or Exception): Type
                of exception.

        Returns:
            Registered function if found, ``None`` otherwise.

        """  # noqa: E501
        exception_class = type(exception)

        for name in (route_name, None):
            exception_key = (exception_class, name)
            handler = self.cached_handlers.get(exception_key)
            if handler:
                return handler

        for name in (route_name, None):
            for ancestor in type.mro(exception_class):
                exception_key = (ancestor, name)
                if exception_key in self.cached_handlers:
                    handler = self.cached_handlers[exception_key]
                    self.cached_handlers[(exception_class, route_name)] = (
                        handler
                    )
                    return handler

                if ancestor is BaseException:
                    break
        self.cached_handlers[(exception_class, route_name)] = None
        handler = None
        return handler

    _lookup = _full_lookup

    def response(self, request, exception):
        """Fetch and executes an exception handler and returns a response object.

        Args:
            request (sanic.request.Request): Instance of the request.
            exception (sanic.exceptions.SanicException or Exception): Exception to handle.

        Returns:
            Wrap the return value obtained from the `default` function or the registered handler for that type of exception.

        """  # noqa: E501
        route_name = request.name if request else None
        handler = self._lookup(exception, route_name)
        response = None
        try:
            if handler:
                response = handler(request, exception)
            if response is None:
                response = self.default(request, exception)
        except Exception:
            try:
                url = repr(request.url)
            except AttributeError:  # no cov
                url = "unknown"
            response_message = (
                'Exception raised in exception handler "%s" for uri: %s'
            )
            error_logger.exception(response_message, handler.__name__, url)

            if self.debug:
                return text(response_message % (handler.__name__, url), 500)
            else:
                return text("An error occurred while handling an error", 500)
        return response

    def default(self, request: Request, exception: Exception) -> HTTPResponse:
        """Provide a default behavior for the objects of ErrorHandler.

        If a developer chooses to extend the ErrorHandler, they can
        provide a custom implementation for this method to behave in a way
        they see fit.

        Args:
            request (sanic.request.Request): Incoming request.
            exception (sanic.exceptions.SanicException or Exception): Exception object.

        Returns:
            HTTPResponse: The response object.

        Examples:
            ```python
            class CustomErrorHandler(ErrorHandler):
                def default(self, request: Request, exception: Exception) -> HTTPResponse:
                    # Custom logic for handling the exception and creating a response
                    custom_response = my_custom_logic(request, exception)
                    return custom_response

            app = Sanic("MyApp", error_handler=CustomErrorHandler())
            ```
        """  # noqa: E501
        self.log(request, exception)
        fallback = request.app.config.FALLBACK_ERROR_FORMAT
        return exception_response(
            request,
            exception,
            debug=self.debug,
            base=self.base,
            fallback=fallback,
        )

    @staticmethod
    def log(request: Request, exception: Exception) -> None:
        """Logs information about an incoming request and the associated exception.

        Args:
            request (Request): The incoming request to be logged.
            exception (Exception): The exception that occurred during the handling of the request.

        Returns:
            None
        """  # noqa: E501
        quiet = getattr(exception, "quiet", False)
        noisy = getattr(request.app.config, "NOISY_EXCEPTIONS", False)
        if quiet is False or noisy is True:
            try:
                url = repr(request.url)
            except AttributeError:  # no cov
                url = "unknown"

            error_logger.exception(
                "Exception occurred while handling uri: %s", url
            )



from __future__ import annotations

from collections.abc import Sequence
from email.utils import formatdate
from functools import partial, wraps
from os import PathLike, path
from pathlib import Path, PurePath
from urllib.parse import unquote

from sanic_routing.route import Route

from sanic.base.meta import SanicMeta
from sanic.compat import clear_function_annotate, stat_async
from sanic.exceptions import FileNotFound, HeaderNotFound, RangeNotSatisfiable
from sanic.handlers import ContentRangeHandler
from sanic.handlers.directory import DirectoryHandler
from sanic.log import error_logger
from sanic.mixins.base import BaseMixin
from sanic.models.futures import FutureStatic
from sanic.request import Request
from sanic.response import HTTPResponse, file, file_stream, validate_file
from sanic.response.convenience import guess_content_type


class StaticMixin(BaseMixin, metaclass=SanicMeta):
    def __init__(self, *args, **kwargs) -> None:
        self._future_statics: set[FutureStatic] = set()

    def _apply_static(self, static: FutureStatic) -> Route:
        raise NotImplementedError  # noqa

    def static(
        self,
        uri: str,
        file_or_directory: PathLike | str,
        pattern: str = r"/?.+",
        use_modified_since: bool = True,
        use_content_range: bool = False,
        stream_large_files: bool | int = False,
        name: str = "static",
        host: str | None = None,
        strict_slashes: bool | None = None,
        content_type: str | None = None,
        apply: bool = True,
        resource_type: str | None = None,
        index: str | Sequence[str] | None = None,
        directory_view: bool = False,
        directory_handler: DirectoryHandler | None = None,
        follow_external_symlink_files: bool = False,
        follow_external_symlink_dirs: bool = False,
    ):
        """Register a root to serve files from. The input can either be a file or a directory.

        This method provides an easy and simple way to set up the route necessary to serve static files.

        Args:
            uri (str): URL path to be used for serving static content.
            file_or_directory (Union[PathLike, str]): Path to the static file
                or directory with static files.
            pattern (str, optional): Regex pattern identifying the valid
                static files. Defaults to `r"/?.+"`.
            use_modified_since (bool, optional): If true, send file modified
                time, and return not modified if the browser's matches the
                server's. Defaults to `True`.
            use_content_range (bool, optional): If true, process header for
                range requests and sends  the file part that is requested.
                Defaults to `False`.
            stream_large_files (Union[bool, int], optional): If `True`, use
                the `StreamingHTTPResponse.file_stream` handler rather than
                the `HTTPResponse.file handler` to send the file. If this
                is an integer, it represents the threshold size to switch
                to `StreamingHTTPResponse.file_stream`. Defaults to `False`,
                which means that the response will not be streamed.
            name (str, optional): User-defined name used for url_for.
                Defaults to `"static"`.
            host (Optional[str], optional): Host IP or FQDN for the
                service to use.
            strict_slashes (Optional[bool], optional): Instruct Sanic to
                check if the request URLs need to terminate with a slash.
            content_type (Optional[str], optional): User-defined content type
                for header.
            apply (bool, optional): If true, will register the route
                immediately. Defaults to `True`.
            resource_type (Optional[str], optional): Explicitly declare a
                resource to be a `"file"` or a `"dir"`.
            index (Optional[Union[str, Sequence[str]]], optional): When
                exposing against a directory, index is  the name that will
                be served as the default file. When multiple file names are
                passed, then they will be tried in order.
            directory_view (bool, optional): Whether to fallback to showing
                the directory viewer when exposing a directory. Defaults
                to `False`.
            directory_handler (Optional[DirectoryHandler], optional): An
                instance of DirectoryHandler that can be used for explicitly
                controlling and subclassing the behavior of the default
                directory handler.
            follow_external_symlink_files (bool, optional): Whether to serve
                files that are symlinks pointing outside the static root.
                Defaults to `False` for security.
            follow_external_symlink_dirs (bool, optional): Whether to serve
                files from directories that are symlinks pointing outside
                the static root. Defaults to `False` for security.

        Returns:
            List[sanic.router.Route]: Routes registered on the router.

        Examples:
            Serving a single file:
            ```python
            app.static('/foo', 'path/to/static/file.txt')
            ```

            Serving all files from a directory:
            ```python
            app.static('/static', 'path/to/static/directory')
            ```

            Serving large files with a specific threshold:
            ```python
            app.static('/static', 'path/to/large/files', stream_large_files=1000000)
            ```
        """  # noqa: E501

        name = self.generate_name(name)

        if strict_slashes is None and self.strict_slashes is not None:
            strict_slashes = self.strict_slashes

        if not isinstance(file_or_directory, (str, bytes, PurePath)):
            raise ValueError(
                f"Static route must be a valid path, not {file_or_directory}"
            )

        try:
            file_or_directory = Path(file_or_directory).resolve()
        except TypeError:
            raise TypeError(
                "Static file or directory must be a path-like object or string"
            )

        if directory_handler and (directory_view or index):
            raise ValueError(
                "When explicitly setting directory_handler, you cannot "
                "set either directory_view or index. Instead, pass "
                "these arguments to your DirectoryHandler instance."
            )

        if not directory_handler:
            directory_handler = DirectoryHandler(
                uri=uri,
                directory=file_or_directory,
                directory_view=directory_view,
                index=index,
                root_path=file_or_directory,
                follow_external_symlink_files=follow_external_symlink_files,
                follow_external_symlink_dirs=follow_external_symlink_dirs,
            )

        static = FutureStatic(
            uri,
            file_or_directory,
            pattern,
            use_modified_since,
            use_content_range,
            stream_large_files,
            name,
            host,
            strict_slashes,
            content_type,
            resource_type,
            directory_handler,
            follow_external_symlink_files,
            follow_external_symlink_dirs,
        )
        self._future_statics.add(static)

        if apply:
            self._apply_static(static)


class StaticHandleMixin(metaclass=SanicMeta):
    def _apply_static(self, static: FutureStatic) -> Route:
        return self._register_static(static)

    def _register_static(
        self,
        static: FutureStatic,
    ):
        # TODO: Though sanic is not a file server, I feel like we should
        # at least make a good effort here.  Modified-since is nice, but
        # we could also look into etags, expires, and caching
        """
        Register a static directory handler with Sanic by adding a route to the
        router and registering a handler.
        """
        file_or_directory: PathLike

        if isinstance(static.file_or_directory, bytes):
            file_or_directory = Path(static.file_or_directory.decode("utf-8"))
        elif isinstance(static.file_or_directory, PurePath):
            file_or_directory = static.file_or_directory
        elif isinstance(static.file_or_directory, str):
            file_or_directory = Path(static.file_or_directory)
        else:
            raise ValueError("Invalid file path string.")

        uri = static.uri
        name = static.name
        # If we're not trying to match a file directly,
        # serve from the folder
        if not static.resource_type:
            if not path.isfile(file_or_directory):
                uri = uri.rstrip("/")
                uri += "/<__file_uri__:path>"
        elif static.resource_type == "dir":
            if path.isfile(file_or_directory):
                raise TypeError(
                    "Resource type improperly identified as directory. "
                    f"'{file_or_directory}'"
                )
            uri = uri.rstrip("/")
            uri += "/<__file_uri__:path>"
        elif static.resource_type == "file" and not path.isfile(
            file_or_directory
        ):
            raise TypeError(
                "Resource type improperly identified as file. "
                f"'{file_or_directory}'"
            )
        elif static.resource_type != "file":
            raise ValueError(
                "The resource_type should be set to 'file' or 'dir'"
            )

        # special prefix for static files
        # if not static.name.startswith("_static_"):
        #     name = f"_static_{static.name}"

        _handler = wraps(self._static_request_handler)(
            partial(
                self._static_request_handler,
                file_or_directory=str(file_or_directory),
                use_modified_since=static.use_modified_since,
                use_content_range=static.use_content_range,
                stream_large_files=static.stream_large_files,
                content_type=static.content_type,
                directory_handler=static.directory_handler,
                follow_external_symlink_files=static.follow_external_symlink_files,
                follow_external_symlink_dirs=static.follow_external_symlink_dirs,
            )
        )

        route, _ = self.route(  # type: ignore
            uri=uri,
            methods=["GET", "HEAD"],
            name=name,
            host=static.host,
            strict_slashes=static.strict_slashes,
            static=True,
        )(_handler)

        return route

    async def _static_request_handler(
        self,
        request: Request,
        *,
        file_or_directory: str,
        use_modified_since: bool,
        use_content_range: bool,
        stream_large_files: bool | int,
        directory_handler: DirectoryHandler,
        follow_external_symlink_files: bool,
        follow_external_symlink_dirs: bool,
        content_type: str | None = None,
        __file_uri__: str | None = None,
    ):
        not_found = FileNotFound(
            "File not found",
            path=Path(file_or_directory),
            relative_url=__file_uri__,
        )

        # Merge served directory and requested file if provided
        file_path = await self._get_file_path(
            file_or_directory,
            __file_uri__,
            not_found,
            follow_external_symlink_files,
            follow_external_symlink_dirs,
        )

        try:
            headers = {}
            # Check if the client has been sent this file before
            # and it has not been modified since
            stats = None
            if use_modified_since:
                stats = await stat_async(file_path)
                modified_since = stats.st_mtime
                response = await validate_file(request.headers, modified_since)
                if response:
                    return response
                headers["Last-Modified"] = formatdate(
                    modified_since, usegmt=True
                )
            _range = None
            if use_content_range:
                _range = None
                if not stats:
                    stats = await stat_async(file_path)
                headers["Accept-Ranges"] = "bytes"
                headers["Content-Length"] = str(stats.st_size)
                if request.method != "HEAD":
                    try:
                        _range = ContentRangeHandler(request, stats)
                    except HeaderNotFound:
                        pass
                    else:
                        del headers["Content-Length"]
                        headers.update(_range.headers)

            if "content-type" not in headers:
                content_type = content_type or guess_content_type(file_path)

                if "charset=" not in content_type and (
                    content_type.startswith("text/")
                    or content_type == "application/javascript"
                ):
                    content_type += "; charset=utf-8"

                headers["Content-Type"] = content_type

            if request.method == "HEAD":
                return HTTPResponse(headers=headers)
            else:
                if stream_large_files:
                    if isinstance(stream_large_files, bool):
                        threshold = 1024 * 1024
                    else:
                        threshold = stream_large_files

                    if not stats:
                        stats = await stat_async(file_path)
                    if stats.st_size >= threshold:
                        return await file_stream(
                            file_path, headers=headers, _range=_range
                        )
                return await file(file_path, headers=headers, _range=_range)
        except (IsADirectoryError, PermissionError):
            return await directory_handler.handle(request, request.path)
        except RangeNotSatisfiable:
            raise
        except FileNotFoundError:
            raise not_found
        except Exception:
            error_logger.exception(
                "Exception in static request handler: "
                f"path={file_or_directory}, "
                f"relative_url={__file_uri__}"
            )
            raise

    async def _get_file_path(
        self,
        file_or_directory,
        __file_uri__,
        not_found,
        follow_external_symlink_files: bool,
        follow_external_symlink_dirs: bool,
    ):
        """
        Resolve a filesystem path safely.

        Security goals:
        - Prevent path traversal via `..`
        - Prevent escaping the root via symlinks unless explicitly allowed
        - Treat file URIs as relative paths even if they look absolute
        """

        def reject():
            error_logger.exception(
                f"File not found: path={file_or_directory}, "
                f"relative_url={__file_uri__}"
            )
            raise not_found

        root_raw = Path(unquote(file_or_directory))
        root_path = root_raw.resolve()
        file_path_raw = root_raw

        if __file_uri__:
            # URLs may start with `/`, Path() interprets as absolute
            rel_uri = unquote(__file_uri__).lstrip("/")
            file_path_raw = Path(root_raw, rel_uri)

            if ".." in file_path_raw.parts:
                reject()

        file_path = file_path_raw.resolve()

        try:
            file_path.relative_to(root_path)
        except ValueError:
            # Check if it's a symlink and determine its type
            is_file_symlink = (
                file_path_raw.is_symlink() and not file_path.is_dir()
            )
            if is_file_symlink:
                allowed = follow_external_symlink_files
            else:
                allowed = follow_external_symlink_dirs
            if not allowed:
                reject()

        return file_path


# Clear __annotate__ on methods that may be pickled via functools.partial
# to avoid PicklingError in Python 3.14+ (PEP 649)
clear_function_annotate(
    StaticHandleMixin._static_request_handler,
    StaticHandleMixin._get_file_path,
    StaticHandleMixin._register_static,
)



from __future__ import annotations

from ast import Assign, Constant, NodeTransformer, Subscript
from typing import TYPE_CHECKING, Any

from sanic.http.constants import HTTP

from .base import BaseScheme


if TYPE_CHECKING:
    from sanic import Sanic


class AltSvcCheck(BaseScheme):
    ident = "ALTSVC"

    def visitors(self) -> list[NodeTransformer]:
        return [RemoveAltSvc(self.app, self.app.state.verbosity)]


class RemoveAltSvc(NodeTransformer):
    def __init__(self, app: Sanic, verbosity: int = 0) -> None:
        self._app = app
        self._verbosity = verbosity
        self._versions = {
            info.settings["version"] for info in app.state.server_info
        }

    def visit_Assign(self, node: Assign) -> Any:
        if any(self._matches(target) for target in node.targets):
            if self._should_remove():
                return None
            assert isinstance(node.value, Constant)
            node.value.value = self.value()
        return node

    def _should_remove(self) -> bool:
        return len(self._versions) == 1

    @staticmethod
    def _matches(node) -> bool:
        return (
            isinstance(node, Subscript)
            and isinstance(node.slice, Constant)
            and node.slice.value == "alt-svc"
        )

    def value(self):
        values = []
        for info in self._app.state.server_info:
            port = info.settings["port"]
            version = info.settings["version"]
            if version is HTTP.VERSION_3:
                values.append(f'h3=":{port}"')
        return ", ".join(values)



from __future__ import annotations

import os
import signal
import sys

from argparse import ArgumentParser
from contextlib import suppress
from pathlib import Path

from sanic.worker.daemon import Daemon


def _add_target_args(parser: ArgumentParser) -> None:
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--pid",
        type=int,
        metavar="PID",
        help="Process ID of the daemon",
    )
    group.add_argument(
        "--pidfile",
        type=str,
        metavar="PATH",
        help="Path to PID file of the daemon",
    )


def make_kill_parser(parser: ArgumentParser) -> None:
    """Kill command always sends SIGKILL."""
    _add_target_args(parser)


def make_status_parser(parser: ArgumentParser) -> None:
    _add_target_args(parser)


def make_restart_parser(parser: ArgumentParser) -> None:
    _add_target_args(parser)


def resolve_target(
    pid: int | None, pidfile: str | None
) -> tuple[int, Path | None]:
    """
    Resolve a PID from either a direct PID or a pidfile path.

    Returns:
        Tuple of (pid, pidfile_path or None)

    Raises:
        SystemExit: If pidfile not found or invalid
    """
    if pid:
        return pid, None

    pidfile_path = Path(pidfile)  # type: ignore[arg-type]
    if not pidfile_path.exists():
        print(f"PID file not found: {pidfile_path}", file=sys.stderr)
        sys.exit(1)

    resolved_pid = Daemon.read_pidfile(pidfile_path)
    if resolved_pid is None:
        print(f"Invalid PID file: {pidfile_path}", file=sys.stderr)
        sys.exit(1)

    return resolved_pid, pidfile_path


def _terminate_process(
    pid: int, sig: signal.Signals, pidfile: Path | None = None
) -> None:
    """Send a signal to terminate a process and clean up pidfile."""
    sig_name = sig.name

    try:
        os.kill(pid, sig)
        print(f"Sent {sig_name} to process {pid}")
    except ProcessLookupError:
        print(f"Process {pid} not found", file=sys.stderr)
        sys.exit(1)
    except PermissionError:
        print(
            f"Permission denied to signal process {pid}. "
            "Are you running as the correct user?",
            file=sys.stderr,
        )
        sys.exit(1)
    except OSError as e:
        print(f"Failed to signal process {pid}: {e}", file=sys.stderr)
        sys.exit(1)

    if pidfile:
        if pidfile.exists():
            try:
                pidfile.unlink()
                print(f"Removed PID file: {pidfile}")
            except OSError as e:
                print(
                    f"Warning: Could not remove PID file {pidfile}: {e}",
                    file=sys.stderr,
                )
        lockfile = pidfile.with_suffix(".lock")
        if lockfile.exists():
            try:
                lockfile.unlink()
            except OSError:
                pass


def kill_daemon(pid: int, pidfile: Path | None = None) -> None:
    """Force kill a daemon process with SIGKILL."""
    _terminate_process(pid, signal.SIGKILL, pidfile)


def stop_daemon(
    pid: int, pidfile: Path | None = None, force: bool = False
) -> None:
    """Stop a daemon process gracefully (SIGTERM) or forcefully (SIGKILL)."""
    sig = signal.SIGKILL if force else signal.SIGTERM
    _terminate_process(pid, sig, pidfile)


def status_daemon(pid: int, pidfile: Path | None = None) -> bool:
    """
    Check if a daemon process is running.

    Args:
        pid: Process ID to check
        pidfile: Optional pidfile path to clean up if stale

    Returns:
        True if running, False otherwise (exits with code 1 if not)
    """
    try:
        os.kill(pid, 0)
        running = True
    except ProcessLookupError:
        running = False
    except PermissionError:
        running = True

    if running:
        print(f"Process {pid} is running")
        return True

    print(f"Process {pid} is NOT running")
    if pidfile and pidfile.exists():
        with suppress(OSError):
            pidfile.unlink()
            print(f"Removed stale PID file: {pidfile}")
    sys.exit(1)


def restart_daemon(pid: int) -> None:
    """
    Restart a daemon process.

    Args:
        pid: Process ID to restart (unused, for future use)
    """
    print("Restart is not yet implemented. Coming in a future release.")
    sys.exit(0)



import asyncio
import os
import platform
import signal
import sys

from collections.abc import Awaitable
from contextlib import contextmanager
from enum import Enum
from typing import Literal

from multidict import CIMultiDict  # type: ignore

from sanic.helpers import Default
from sanic.log import error_logger


StartMethod = (
    Default | Literal["fork"] | Literal["forkserver"] | Literal["spawn"]
)

OS_IS_WINDOWS = os.name == "nt"
PYPY_IMPLEMENTATION = platform.python_implementation() == "PyPy"
UVLOOP_INSTALLED = False
PYTHON_314_OR_LATER = sys.version_info >= (3, 14)

try:
    import uvloop  # type: ignore # noqa

    UVLOOP_INSTALLED = True
except ImportError:
    pass

# Python 3.11 changed the way Enum formatting works for mixed-in types.
if sys.version_info < (3, 11, 0):

    class StrEnum(str, Enum):
        pass

else:
    from enum import StrEnum  # type: ignore # noqa


class UpperStrEnum(StrEnum):
    """Base class for string enums that are case insensitive."""

    def _generate_next_value_(name, start, count, last_values):
        return name.upper()

    def __eq__(self, value: object) -> bool:
        value = str(value).upper()
        return super().__eq__(value)

    def __hash__(self) -> int:
        return hash(self.value)

    def __str__(self) -> str:
        return self.value


@contextmanager
def use_context(method: StartMethod):
    from sanic import Sanic

    orig = Sanic.start_method
    Sanic.start_method = method
    yield
    Sanic.start_method = orig


def enable_windows_color_support():
    import ctypes

    kernel = ctypes.windll.kernel32
    kernel.SetConsoleMode(kernel.GetStdHandle(-11), 7)


def pypy_os_module_patch() -> None:
    """
    The PyPy os module is missing the 'readlink' function, which causes issues
    withaiofiles. This workaround replaces the missing 'readlink' function
    with 'os.path.realpath', which serves the same purpose.
    """
    if hasattr(os, "readlink"):
        error_logger.debug(
            "PyPy: Skipping patching of the os module as it appears the "
            "'readlink' function has been added."
        )
        return

    module = sys.modules["os"]
    module.readlink = os.path.realpath  # type: ignore


def pypy_windows_set_console_cp_patch() -> None:
    """
    A patch function for PyPy on Windows that sets the console code page to
    UTF-8 encodingto allow for proper handling of non-ASCII characters. This
    function uses ctypes to call the Windows API functions SetConsoleCP and
    SetConsoleOutputCP to set the code page.
    """
    from ctypes import windll  # type: ignore

    code: int = windll.kernel32.GetConsoleOutputCP()
    if code != 65001:
        windll.kernel32.SetConsoleCP(65001)
        windll.kernel32.SetConsoleOutputCP(65001)


class Header(CIMultiDict):
    """Container used for both request and response headers.
    It is a subclass of  [CIMultiDict](https://multidict.readthedocs.io/en/stable/multidict.html#cimultidictproxy)

    It allows for multiple values for a single key in keeping with the HTTP
    spec. Also, all keys are *case in-sensitive*.

    Please checkout [the MultiDict documentation](https://multidict.readthedocs.io/en/stable/multidict.html#multidict)
    for more details about how to use the object. In general, it should work
    very similar to a regular dictionary.
    """  # noqa: E501

    def __getattr__(self, key: str) -> str:
        if key.startswith("_"):
            return self.__getattribute__(key)
        key = key.rstrip("_").replace("_", "-")
        return ",".join(self.getall(key, []))

    def get_all(self, key: str):
        """Convenience method mapped to ``getall()``."""
        return self.getall(key, [])


use_trio = sys.argv[0].endswith("hypercorn") and "trio" in sys.argv

if use_trio:  # pragma: no cover
    import trio  # type: ignore

    def stat_async(path) -> Awaitable[os.stat_result]:
        return trio.Path(path).stat()

    open_async = trio.open_file
    CancelledErrors = tuple([asyncio.CancelledError, trio.Cancelled])
else:
    if PYPY_IMPLEMENTATION:
        pypy_os_module_patch()

        if OS_IS_WINDOWS:
            pypy_windows_set_console_cp_patch()

    from aiofiles import open as aio_open  # type: ignore
    from aiofiles.os import stat as stat_async  # type: ignore  # noqa: F401

    async def open_async(file, mode="r", **kwargs):
        return aio_open(file, mode, **kwargs)

    CancelledErrors = tuple([asyncio.CancelledError])


def ctrlc_workaround_for_windows(app):
    async def stay_active(app):
        """Asyncio wakeups to allow receiving SIGINT in Python"""
        while not die:
            # If someone else stopped the app, just exit
            if app.state.is_stopping:
                return
            # Windows Python blocks signal handlers while the event loop is
            # waiting for I/O. Frequent wakeups keep interrupts flowing.
            await asyncio.sleep(0.1)
        # Can't be called from signal handler, so call it from here
        app.stop()

    def ctrlc_handler(sig, frame):
        nonlocal die
        if die:
            raise KeyboardInterrupt("Non-graceful Ctrl+C")
        die = True

    die = False
    signal.signal(signal.SIGINT, ctrlc_handler)
    app.add_task(stay_active)


def clear_function_annotate(*funcs):
    """Clear __annotate__ on functions for Python 3.14+ pickle compatibility.

    In Python 3.14, PEP 649 adds __annotate__ to functions with annotations.
    When methods are used in functools.partial and pickled, the __annotate__
    function can cause PicklingError because pickle cannot locate it by name.

    This function sets __annotate__ to None on the given functions to avoid
    pickle issues.
    """
    if PYTHON_314_OR_LATER:
        for func in funcs:
            if hasattr(func, "__annotate__") and func.__annotate__ is not None:
                func.__annotate__ = None



from __future__ import annotations

import atexit
import grp
import os
import pwd
import signal
import sys
import time
import uuid

from dataclasses import dataclass
from pathlib import Path

from sanic.compat import OS_IS_WINDOWS
from sanic.log import logger


try:
    import fcntl  # type: ignore
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    fcntl = None  # type: ignore


def _get_default_runtime_dir() -> Path:
    """
    Default directory for auto-generated runtime artifacts (pid/lock/log).

    Priority:
      1) XDG_RUNTIME_DIR/sanic  (preferred, runtime-only)
      2) ~/.local/state/sanic   (persistent state, modern default)
      3) ~/.cache/sanic         (fallback)
      4) cwd                    (last resort)
    """
    xdg_runtime = os.environ.get("XDG_RUNTIME_DIR")
    if xdg_runtime:
        path = Path(xdg_runtime) / "sanic"
        try:
            path.mkdir(mode=0o700, parents=True, exist_ok=True)
            return path
        except OSError:
            pass

    state_dir = Path.home() / ".local" / "state" / "sanic"
    try:
        state_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
        return state_dir
    except OSError:
        pass

    cache_dir = Path.home() / ".cache" / "sanic"
    try:
        cache_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
        return cache_dir
    except OSError:
        pass

    return Path.cwd()


def _sanitize_name(name: str) -> str:
    return (
        "".join(
            c if c.isalnum() or c in ("-", "_", ".") else "_" for c in name
        ).strip("._")
        or "sanic"
    )


def _process_exists(pid: int) -> bool:
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True


def _is_sanic_process(pid: int) -> bool:
    """
    Best-effort Sanic process identification.

    On Linux, checks /proc/<pid>/cmdline for 'sanic'. On other platforms,
    falls back to True if process exists (no strong identification).
    """
    proc_cmdline = Path(f"/proc/{pid}/cmdline")
    if proc_cmdline.exists():
        try:
            data = proc_cmdline.read_bytes()
            return b"sanic" in data
        except OSError:
            return False
    return True


@dataclass(frozen=True)
class PidfileInfo:
    pid: int
    started: int | None = None
    name: str | None = None


class DaemonError(Exception):
    pass


class Daemon:
    """
    Daemonization helper (Unix only).

    Supports:
      - Double-fork daemonization
      - Optional PID file (with Sanic marker + metadata)
      - Optional lock file (prevents double start / stale PID reuse issues)
      - Optional logfile redirection
      - Optional privilege drop (user/group)
      - SIGHUP handler that preserves pidfile identity
    """

    def __init__(
        self,
        pidfile: str | None = None,
        logfile: str | None = None,
        user: str | None = None,
        group: str | None = None,
        name: str | None = None,
        lockfile: str | None = None,
    ):
        if OS_IS_WINDOWS:
            raise DaemonError(
                "Daemon mode is not supported on Windows. "
                "Consider using a Windows service or nssm instead."
            )

        self.name = _sanitize_name(name) if name else None

        self._auto_pidfile = pidfile == "auto" or pidfile == ""
        self.pidfile: Path | None
        if self._auto_pidfile:
            base = _get_default_runtime_dir()
            # Prefer deterministic name if provided
            if self.name:
                self.pidfile = base / f"{self.name}.pid"
            else:
                self.pidfile = base / f"sanic-{uuid.uuid4().hex}.pid"
        elif pidfile:
            self.pidfile = Path(pidfile)
        else:
            self.pidfile = None

        self.logfile = Path(logfile) if logfile else None
        self.user = user
        self.group = group

        self._uid: int | None = None
        self._gid: int | None = None
        self.pid: int | None = None

        self._lockfile_path: Path | None = Path(lockfile) if lockfile else None
        self._lock_fd: int | None = None

    @staticmethod
    def get_pidfile_path(name: str) -> Path:
        """
        Get the predictable pidfile path for a given name.

        Args:
            name: The application/daemon name

        Returns:
            Path to the pidfile in the default runtime directory
        """
        sanitized = _sanitize_name(name)
        return _get_default_runtime_dir() / f"{sanitized}.pid"

    def validate(self) -> None:
        self._validate_user_group()
        self._validate_paths()
        self._validate_runtime_state()

    def daemonize(self) -> None:
        """
        Double-fork daemonization.

        Important: anything meant for the invoking terminal must be
        printed/logged before calling this method, since stdout/stderr
        are redirected after detaching.
        """
        self.validate()

        # Acquire lock before forking to prevent race condition
        # The lock is inherited by child processes and remains held
        self._acquire_lockfile()

        # First fork
        try:
            pid = os.fork()
            if pid > 0:
                os._exit(0)  # pragma: no cover
        except OSError as e:
            raise DaemonError(f"First fork failed: {e}") from e

        os.chdir("/")
        os.setsid()
        os.umask(0o022)

        # Second fork
        try:
            pid = os.fork()
            if pid > 0:
                os._exit(0)  # pragma: no cover
        except OSError as e:
            raise DaemonError(f"Second fork failed: {e}") from e

        self.pid = os.getpid()

        self._redirect_streams()
        self._write_pidfile()
        self._setup_signal_handlers()

        logger.info(
            "Sanic daemon started",
            extra={
                "daemon_pid": self.pid,
                "daemon_pidfile": str(self.pidfile) if self.pidfile else None,
                "daemon_logfile": str(self.logfile) if self.logfile else None,
                "daemon_name": self.name,
            },
        )

    def drop_privileges(self) -> None:
        """
        Drop privileges to configured user/group.

        Call after binding privileged ports but before serving requests.
        """
        if self._uid is None and self._gid is None:
            return

        if os.getuid() != 0:
            logger.warning(
                "Privilege drop requested but not running as root",
                extra={"user": self.user, "group": self.group},
            )
            return

        if self._gid is not None:
            try:
                os.setgid(self._gid)
                if self.user:
                    os.initgroups(self.user, self._gid)
            except PermissionError as e:
                grp = self.group or self._gid
                raise DaemonError(
                    f"Cannot change group to '{grp}'. Are you root?"
                ) from e

        if self._uid is not None:
            try:
                os.setuid(self._uid)
            except PermissionError as e:
                usr = self.user or self._uid
                raise DaemonError(
                    f"Cannot change user to '{usr}'. Are you root?"
                ) from e

        logger.info(
            "Dropped privileges",
            extra={
                "user": self.user or "unchanged",
                "group": self.group or "unchanged",
            },
        )

    def _validate_user_group(self) -> None:
        if not self.user and not self.group:
            return

        if self.user:
            try:
                pw = pwd.getpwnam(self.user)
            except KeyError as e:
                raise DaemonError(f"User '{self.user}' does not exist") from e
            self._uid = pw.pw_uid
            if not self.group:
                self._gid = pw.pw_gid

        if self.group:
            try:
                gr = grp.getgrnam(self.group)
            except KeyError as e:
                raise DaemonError(
                    f"Group '{self.group}' does not exist"
                ) from e
            self._gid = gr.gr_gid

    def _validate_paths(self) -> None:
        if self.pidfile:
            self._validate_writable_dir(self.pidfile, "PID file")

        if self.logfile:
            self._validate_writable_dir(self.logfile, "Log file")

        # Default lockfile beside pidfile if not explicitly set
        if not self._lockfile_path and self.pidfile:
            self._lockfile_path = self.pidfile.with_suffix(".lock")

        if self._lockfile_path:
            self._validate_writable_dir(self._lockfile_path, "Lock file")

    def _validate_runtime_state(self) -> None:
        # PID file check (stale vs running)
        if not self.pidfile or not self.pidfile.exists():
            return

        info = self.read_pidfile_info(self.pidfile)
        if not info:
            return

        if _process_exists(info.pid) and _is_sanic_process(info.pid):
            raise DaemonError(f"Daemon already running with PID {info.pid}")

    def _validate_writable_dir(self, path: Path, label: str) -> None:
        directory = path.parent
        if not directory.exists():
            raise DaemonError(f"{label} directory does not exist: {directory}")
        if not os.access(directory, os.W_OK):
            raise DaemonError(
                f"Cannot write to {label} directory: {directory}"
            )

    def _write_pidfile(self) -> None:
        if not self.pidfile:
            return

        pid = os.getpid()
        started = int(time.time())
        name = self.name or ""

        # Format:
        #   sanic
        #   pid=123
        #   started=...
        #   name=...
        lines = [
            "sanic",
            f"pid={pid}",
            f"started={started}",
        ]
        if name:
            lines.append(f"name={name}")

        try:
            self.pidfile.write_text("\n".join(lines) + "\n")
        except OSError as e:
            raise DaemonError(
                f"Failed to write PID file: {self.pidfile}"
            ) from e

        atexit.register(self._remove_pidfile)

    def _remove_pidfile(self) -> None:
        if not self.pidfile:
            return
        try:
            self.pidfile.unlink(missing_ok=True)
        except OSError as e:
            # Best-effort cleanup: failure to remove the PID file is non-fatal.
            logger.debug("Failed to remove PID file %s: %s", self.pidfile, e)

    @staticmethod
    def read_pidfile(pidfile: str | Path) -> int | None:
        info = Daemon.read_pidfile_info(pidfile)
        return info.pid if info else None

    @staticmethod
    def read_pidfile_info(pidfile: str | Path) -> PidfileInfo | None:
        path = Path(pidfile)
        if not path.exists():
            return None

        try:
            raw = path.read_text().splitlines()
        except OSError:
            return None

        if not raw or raw[0].strip() != "sanic":
            return None

        data: dict[str, str] = {}
        for line in raw[1:]:
            if "=" in line:
                k, v = line.split("=", 1)
                data[k.strip()] = v.strip()

        try:
            pid = int(data.get("pid", ""))
        except ValueError:
            return None

        started = None
        if "started" in data:
            try:
                started = int(data["started"])
            except ValueError:
                started = None

        name = data.get("name") or None
        return PidfileInfo(pid=pid, started=started, name=name)

    def _acquire_lockfile(self) -> None:
        if not self._lockfile_path:
            return
        if fcntl is None:
            return

        try:
            fd = os.open(
                str(self._lockfile_path), os.O_RDWR | os.O_CREAT, 0o644
            )
        except OSError as e:
            raise DaemonError(
                f"Failed to open lock file: {self._lockfile_path}"
            ) from e

        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except OSError as e:
            try:
                os.close(fd)
            except OSError:
                pass
            raise DaemonError(
                f"Daemon already running (lock held): {self._lockfile_path}"
            ) from e

        self._lock_fd = fd
        atexit.register(self._release_lockfile)

    def _release_lockfile(self) -> None:
        if self._lock_fd is None:
            return
        try:
            if fcntl is not None:
                fcntl.flock(self._lock_fd, fcntl.LOCK_UN)
        except OSError as e:
            # Best-effort cleanup: failure to unlock is non-fatal.
            logger.debug(
                "Failed to unlock file descriptor %s: %s", self._lock_fd, e
            )
        try:
            os.close(self._lock_fd)
        except OSError as e:
            # Best-effort cleanup: failure to close is non-fatal.
            logger.debug(
                "Failed to close file descriptor %s: %s", self._lock_fd, e
            )
        self._lock_fd = None
        if self._lockfile_path:
            try:
                self._lockfile_path.unlink(missing_ok=True)
            except OSError as e:
                # Best-effort cleanup: failure to remove lock file
                # is non-fatal.
                logger.debug(
                    "Failed to remove lock file %s: %s", self._lockfile_path, e
                )

    def _redirect_streams(self) -> None:
        sys.stdout.flush()
        sys.stderr.flush()

        stdin_fd = os.open(os.devnull, os.O_RDONLY)
        os.dup2(stdin_fd, sys.stdin.fileno())

        if self.logfile:
            log_fd = os.open(
                str(self.logfile),
                os.O_WRONLY | os.O_CREAT | os.O_APPEND,
                0o644,
            )
            os.dup2(log_fd, sys.stdout.fileno())
            os.dup2(log_fd, sys.stderr.fileno())
            if log_fd > 2:
                os.close(log_fd)
        else:
            devnull_fd = os.open(os.devnull, os.O_RDWR)
            os.dup2(devnull_fd, sys.stdout.fileno())
            os.dup2(devnull_fd, sys.stderr.fileno())
            if devnull_fd > 2:
                os.close(devnull_fd)

        if stdin_fd > 2:
            os.close(stdin_fd)

    def _setup_signal_handlers(self) -> None:
        if not self.pidfile:
            return

        original = signal.getsignal(signal.SIGHUP)

        def _handle_sighup(signum, frame):
            # Preserve pidfile identity; ensure it exists.
            self._write_pidfile()
            if callable(original):
                original(signum, frame)

        signal.signal(signal.SIGHUP, _handle_sighup)



from __future__ import annotations

from typing import TYPE_CHECKING

from sanic.http.constants import Stage


if TYPE_CHECKING:
    from sanic.response import BaseHTTPResponse
    from sanic.server.protocols.http_protocol import HttpProtocol


class Stream:
    stage: Stage
    response: BaseHTTPResponse | None
    protocol: HttpProtocol
    url: str | None
    request_body: bytes | None
    request_max_size: int | float

    __touchup__: tuple[str, ...] = tuple()
    __slots__ = ("request_max_size",)

    def respond(
        self, response: BaseHTTPResponse
    ) -> BaseHTTPResponse:  # no cov
        raise NotImplementedError("Not implemented")



import logging

from sanic.logging.filter import VerbosityFilter


_verbosity_filter = VerbosityFilter()

logger = logging.getLogger("sanic.root")  # no cov
"""
General Sanic logger
"""
logger.addFilter(_verbosity_filter)

error_logger = logging.getLogger("sanic.error")  # no cov
"""
Logger used by Sanic for error logging
"""
error_logger.addFilter(_verbosity_filter)

access_logger = logging.getLogger("sanic.access")  # no cov
"""
Logger used by Sanic for access logging
"""
access_logger.addFilter(_verbosity_filter)

server_logger = logging.getLogger("sanic.server")  # no cov
"""
Logger used by Sanic for server related messages
"""
server_logger.addFilter(_verbosity_filter)

websockets_logger = logging.getLogger("sanic.websockets")  # no cov
"""
Logger used by Sanic for websockets module and protocol related messages
"""
websockets_logger.addFilter(_verbosity_filter)
websockets_logger.setLevel(logging.WARNING)  # Too noisy on debug/info



import errno
import sys

from typing import Callable

from sanic.exceptions import ServerError
from sanic.log import error_logger
from sanic.worker.daemon import DaemonError


ExceptionHandler = Callable[[Exception], bool]


def maybe_handle_startup_error(exc: Exception) -> None:
    for handler in EXCEPTION_HANDLERS:
        if handler(exc):
            sys.exit(1)
    raise exc


def _handle_os_error(exc: Exception) -> bool:
    if not isinstance(exc, OSError):
        return False

    if exc.errno == errno.EADDRINUSE:
        error_logger.error(
            "Startup failed: Address already in use. \n\n"
            "Ensure no other process is using the same address and port, "
            "or configure the server to use a different port."
        )
    else:
        error_logger.error(
            "Startup failed due to OS error: %s (errno %s)",
            exc.strerror,
            exc.errno,
        )
    return True


def _handle_server_error(exc: Exception) -> bool:
    if not isinstance(exc, ServerError):
        return False

    error_logger.error(f"Startup failed due to server error. {exc}")
    return True


def _handle_daemon_error(exc: Exception) -> bool:
    if not isinstance(exc, DaemonError):
        return False

    error_logger.error(f"Daemon error: {exc}")
    return True


EXCEPTION_HANDLERS: tuple[ExceptionHandler, ...] = (
    _handle_os_error,
    _handle_server_error,
    _handle_daemon_error,
)



from __future__ import annotations

import re

from collections.abc import Iterable
from typing import Any
from urllib.parse import unquote

from sanic.exceptions import InvalidHeader
from sanic.helpers import STATUS_CODES


# TODO:
# - the Options object should be a typed object to allow for less casting
#   across the application (in request.py for example)
HeaderIterable = Iterable[tuple[str, Any]]  # Values convertible to str
HeaderBytesIterable = Iterable[tuple[bytes, bytes]]
Options = dict[str, int | str]  # key=value fields in various headers
OptionsIterable = Iterable[tuple[str, str]]  # May contain duplicate keys

_token, _quoted = r"([\w!#$%&'*+\-.^_`|~]+)", r'"([^"]*)"'
_param = re.compile(rf";\s*{_token}=(?:{_token}|{_quoted})", re.ASCII)
_ipv6 = "(?:[0-9A-Fa-f]{0,4}:){2,7}[0-9A-Fa-f]{0,4}"
_ipv6_re = re.compile(_ipv6)
_host_re = re.compile(
    r"((?:\[" + _ipv6 + r"\])|[a-zA-Z0-9.\-]{1,253})(?::(\d{1,5}))?"
)

# RFC's quoted-pair escapes are mostly ignored by browsers. Chrome, Firefox and
# curl all have different escaping, that we try to handle as well as possible,
# even though no client escapes in a way that would allow perfect handling.

# For more information, consult ../tests/test_requests.py


class MediaType:
    """A media type, as used in the Accept header.

    This class is a representation of a media type, as used in the Accept
    header. It encapsulates the type, subtype and any parameters, and
    provides methods for matching against other media types.

    Two separate methods are provided for searching the list:
    - 'match' for finding the most preferred match (wildcards supported)
    -  operator 'in' for checking explicit matches (wildcards as literals)

    Args:
        type_ (str): The type of the media type.
        subtype (str): The subtype of the media type.
        **params (str): Any parameters for the media type.
    """

    def __init__(
        self,
        type_: str,
        subtype: str,
        **params: str,
    ):
        self.type = type_
        self.subtype = subtype
        self.q = float(params.get("q", "1.0"))
        self.params = params
        self.mime = f"{type_}/{subtype}"
        self.key = (
            -1 * self.q,
            -1 * len(self.params),
            self.subtype == "*",
            self.type == "*",
        )

    def __repr__(self):
        return self.mime + "".join(f";{k}={v}" for k, v in self.params.items())

    def __eq__(self, other):
        """Check for mime (str or MediaType) identical type/subtype.
        Parameters such as q are not considered."""
        if isinstance(other, str):
            # Give a friendly reminder if str contains parameters
            if ";" in other:
                raise ValueError("Use match() to compare with parameters")
            return self.mime == other
        if isinstance(other, MediaType):
            # Ignore parameters silently with MediaType objects
            return self.mime == other.mime
        return NotImplemented

    def match(
        self,
        mime_with_params: str | MediaType,
    ) -> MediaType | None:
        """Match this media type against another media type.

        Check if this media type matches the given mime type/subtype.
        Wildcards are supported both ways on both type and subtype.
        If mime contains a semicolon, optionally followed by parameters,
        the parameters of the two media types must match exactly.

        .. note::
            Use the `==` operator instead to check for literal matches
            without expanding wildcards.


        Args:
            media_type (str): A type/subtype string to match.

        Returns:
            MediaType: Returns `self` if the media types are compatible.
            None: Returns `None` if the media types are not compatible.
        """
        mt = (
            MediaType._parse(mime_with_params)
            if isinstance(mime_with_params, str)
            else mime_with_params
        )
        return (
            self
            if (
                mt
                # All parameters given in the other media type must match
                and all(self.params.get(k) == v for k, v in mt.params.items())
                # Subtype match
                and (
                    self.subtype == mt.subtype
                    or self.subtype == "*"
                    or mt.subtype == "*"
                )
                # Type match
                and (
                    self.type == mt.type or self.type == "*" or mt.type == "*"
                )
            )
            else None
        )

    @property
    def has_wildcard(self) -> bool:
        """Return True if this media type has a wildcard in it.

        Returns:
            bool: True if this media type has a wildcard in it.
        """
        return any(part == "*" for part in (self.subtype, self.type))

    @classmethod
    def _parse(cls, mime_with_params: str) -> MediaType | None:
        mtype = mime_with_params.strip()
        if "/" not in mime_with_params:
            return None

        mime, *raw_params = mtype.split(";")
        type_, subtype = mime.split("/", 1)
        if not type_ or not subtype:
            raise ValueError(f"Invalid media type: {mtype}")

        params = {
            key.strip(): value.strip()
            for key, value in (param.split("=", 1) for param in raw_params)
        }

        return cls(type_.lstrip(), subtype.rstrip(), **params)


class Matched:
    """A matching result of a MIME string against a header.

    This class is a representation of a matching result of a MIME string
    against a header. It encapsulates the MIME string, the header, and
    provides methods for matching against other MIME strings.

    Args:
        mime (str): The MIME string to match.
        header (MediaType): The header to match against, if any.
    """

    def __init__(self, mime: str, header: MediaType | None):
        self.mime = mime
        self.header = header

    def __repr__(self):
        return f"<{self} matched {self.header}>" if self else "<no match>"

    def __str__(self):
        return self.mime

    def __bool__(self):
        return self.header is not None

    def __eq__(self, other: Any) -> bool:
        try:
            comp, other_accept = self._compare(other)
        except TypeError:
            return False

        return bool(
            comp
            and (
                (self.header and other_accept.header)
                or (not self.header and not other_accept.header)
            )
        )

    def _compare(self, other) -> tuple[bool, Matched]:
        if isinstance(other, str):
            parsed = Matched.parse(other)
            if self.mime == other:
                return True, parsed
            other = parsed

        if isinstance(other, Matched):
            return self.header == other.header, other

        raise TypeError(
            "Comparison not supported between unequal "
            f"mime types of '{self.mime}' and '{other}'"
        )

    def match(self, other: str | Matched) -> Matched | None:
        """Match this MIME string against another MIME string.

        Check if this MIME string matches the given MIME string. Wildcards are supported both ways on both type and subtype.

        Args:
            other (str): A MIME string to match.

        Returns:
            Matched: Returns `self` if the MIME strings are compatible.
            None: Returns `None` if the MIME strings are not compatible.
        """  # noqa: E501
        accept = Matched.parse(other) if isinstance(other, str) else other
        if not self.header or not accept.header:
            return None
        if self.header.match(accept.header):
            return accept
        return None

    @classmethod
    def parse(cls, raw: str) -> Matched:
        media_type = MediaType._parse(raw)
        return cls(raw, media_type)


class AcceptList(list):
    """A list of media types, as used in the Accept header.

    The Accept header entries are listed in order of preference, starting
    with the most preferred. This class is a list of `MediaType` objects,
    that encapsulate also the q value or any other parameters.

    Two separate methods are provided for searching the list:
    - 'match' for finding the most preferred match (wildcards supported)
    -  operator 'in' for checking explicit matches (wildcards as literals)

    Args:
        *args (MediaType): Any number of MediaType objects.
    """

    def match(self, *mimes: str, accept_wildcards=True) -> Matched:
        """Find a media type accepted by the client.

        This method can be used to find which of the media types requested by
        the client is most preferred against the ones given as arguments.

        The ordering of preference is set by:
        1. The order set by RFC 7231, s. 5.3.2, giving a higher priority
            to q values and more specific type definitions,
        2. The order of the arguments (first is most preferred), and
        3. The first matching entry on the Accept header.

        Wildcards are matched both ways. A match is usually found, as the
        Accept headers typically include `*/*`, in particular if the header
        is missing, is not manually set, or if the client is a browser.

        Note: the returned object behaves as a string of the mime argument
        that matched, and is empty/falsy if no match was found. The matched
        header entry `MediaType` or `None` is available as the `m` attribute.

        Args:
            mimes (List[str]): Any MIME types to search for in order of preference.
            accept_wildcards (bool): Match Accept entries with wildcards in them.

        Returns:
            Match: A match object with the mime string and the MediaType object.
        """  # noqa: E501
        a = sorted(
            (-acc.q, i, j, mime, acc)
            for j, acc in enumerate(self)
            if accept_wildcards or not acc.has_wildcard
            for i, mime in enumerate(mimes)
            if acc.match(mime)
        )
        return Matched(*(a[0][-2:] if a else ("", None)))

    def __str__(self):
        """Format as Accept header value (parsed, not original)."""
        return ", ".join(str(m) for m in self)


def parse_accept(accept: str | None) -> AcceptList:
    """Parse an Accept header and order the acceptable media types according to RFC 7231, s. 5.3.2

    https://datatracker.ietf.org/doc/html/rfc7231#section-5.3.2

    Args:
        accept (str): The Accept header value to parse.

    Returns:
        AcceptList: A list of MediaType objects, ordered by preference.

    Raises:
        InvalidHeader: If the header value is invalid.
    """  # noqa: E501
    if not accept:
        if accept == "":
            return AcceptList()  # Empty header, accept nothing
        accept = "*/*"  # No header means that all types are accepted
    try:
        a = [
            mt
            for mt in [MediaType._parse(mtype) for mtype in accept.split(",")]
            if mt
        ]
        if not a:
            raise ValueError
        return AcceptList(sorted(a, key=lambda x: x.key))
    except ValueError:
        raise InvalidHeader(f"Invalid header value in Accept: {accept}")


def parse_content_header(value: str) -> tuple[str, Options]:
    """Parse content-type and content-disposition header values.

    E.g. `form-data; name=upload; filename="file.txt"` to
    ('form-data', {'name': 'upload', 'filename': 'file.txt'})

    Mostly identical to cgi.parse_header and werkzeug.parse_options_header
    but runs faster and handles special characters better.

    Unescapes %22 to `"` and %0D%0A to `\n` in field values.

    Args:
        value (str): The header value to parse.

    Returns:
        Tuple[str, Options]: The header value and a dict of options.
    """
    pos = value.find(";")
    if pos == -1:
        options: dict[str, int | str] = {}
    else:
        options = {
            m.group(1).lower(): (m.group(2) or m.group(3))
            .replace("%22", '"')
            .replace("%0D%0A", "\n")
            for m in _param.finditer(value[pos:])
        }
        value = value[:pos]
    return value.strip().lower(), options


# https://tools.ietf.org/html/rfc7230#section-3.2.6 and
# https://tools.ietf.org/html/rfc7239#section-4
# This regex is for *reversed* strings because that works much faster for
# right-to-left matching than the other way around. Be wary that all things are
# a bit backwards! _rparam matches forwarded pairs alike ";key=value"
_rparam = re.compile(f"(?:{_token}|{_quoted})={_token}\\s*($|[;,])", re.ASCII)


def parse_forwarded(headers, config) -> Options | None:
    """Parse RFC 7239 Forwarded headers.
    The value of `by` or `secret` must match `config.FORWARDED_SECRET`
    :return: dict with keys and values, or None if nothing matched
    """
    header = headers.getall("forwarded", None)
    secret = config.FORWARDED_SECRET
    if header is None or not secret:
        return None
    header = ",".join(header)  # Join multiple header lines
    if secret not in header:
        return None
    # Loop over <separator><key>=<value> elements from right to left
    sep = pos = None
    options_list: list[tuple[str, str]] = []
    found = False
    for m in _rparam.finditer(header[::-1]):
        # Start of new element? (on parser skips and non-semicolon right sep)
        if m.start() != pos or sep != ";":
            # Was the previous element (from right) what we wanted?
            if found:
                break
            # Clear values and parse as new element
            del options_list[:]
        pos = m.end()
        val_token, val_quoted, key, sep = m.groups()
        key = key.lower()[::-1]
        val = (val_token or val_quoted.replace('"\\', '"'))[::-1]
        options_list.append((key, val))
        if key in ("secret", "by") and val == secret:
            found = True
        # Check if we would return on next round, to avoid useless parse
        if found and sep != ";":
            break
    # If secret was found, return the matching options in left-to-right order
    return fwd_normalize(reversed(options_list)) if found else None


def parse_xforwarded(headers, config) -> Options | None:
    """Parse traditional proxy headers."""
    real_ip_header = config.REAL_IP_HEADER
    proxies_count = config.PROXIES_COUNT
    addr = real_ip_header and headers.getone(real_ip_header, None)
    if not addr and proxies_count:
        assert proxies_count > 0
        try:
            # Combine, split and filter multiple headers' entries
            forwarded_for = headers.getall(config.FORWARDED_FOR_HEADER)
            proxies = [
                p
                for p in (
                    p.strip() for h in forwarded_for for p in h.split(",")
                )
                if p
            ]
            addr = proxies[-proxies_count]
        except (KeyError, IndexError):
            pass
    # No processing of other headers if no address is found
    if not addr:
        return None

    def options():
        yield "for", addr
        for key, header in (
            ("proto", "x-scheme"),
            ("proto", "x-forwarded-proto"),  # Overrides X-Scheme if present
            ("host", "x-forwarded-host"),
            ("port", "x-forwarded-port"),
            ("path", "x-forwarded-path"),
        ):
            yield key, headers.getone(header, None)

    return fwd_normalize(options())


def fwd_normalize(fwd: OptionsIterable) -> Options:
    """Normalize and convert values extracted from forwarded headers.

    Args:
        fwd (OptionsIterable): An iterable of key-value pairs.

    Returns:
        Options: A dict of normalized key-value pairs.
    """
    ret: dict[str, int | str] = {}
    for key, val in fwd:
        if val is not None:
            try:
                if key in ("by", "for"):
                    ret[key] = fwd_normalize_address(val)
                elif key in ("host", "proto"):
                    ret[key] = val.lower()
                elif key == "port":
                    ret[key] = int(val)
                elif key == "path":
                    ret[key] = unquote(val)
                else:
                    ret[key] = val
            except ValueError:
                pass
    return ret


def fwd_normalize_address(addr: str) -> str:
    """Normalize address fields of proxy headers.

    Args:
        addr (str): An address string.

    Returns:
        str: A normalized address string.
    """
    if addr == "unknown":
        raise ValueError()  # omit unknown value identifiers
    if addr.startswith("_"):
        return addr  # do not lower-case obfuscated strings
    if _ipv6_re.fullmatch(addr):
        addr = f"[{addr}]"  # bracket IPv6
    return addr.lower()


def parse_host(host: str) -> tuple[str | None, int | None]:
    """Split host:port into hostname and port.

    Args:
        host (str): A host string.

    Returns:
        Tuple[Optional[str], Optional[int]]: A tuple of hostname and port.
    """
    m = _host_re.fullmatch(host)
    if not m:
        return None, None
    host, port = m.groups()
    return host.lower(), int(port) if port is not None else None


_HTTP1_STATUSLINES = [
    b"HTTP/1.1 %d %b\r\n" % (status, STATUS_CODES.get(status, b"UNKNOWN"))
    for status in range(1000)
]


def format_http1_response(status: int, headers: HeaderBytesIterable) -> bytes:
    """Format a HTTP/1.1 response header.

    Args:
        status (int): The HTTP status code.
        headers (HeaderBytesIterable): An iterable of header tuples.

    Returns:
        bytes: The formatted response header.
    """
    # Note: benchmarks show that here bytes concat is faster than bytearray,
    # b"".join() or %-formatting. %timeit any changes you make.
    ret = _HTTP1_STATUSLINES[status]
    for h in headers:
        ret += b"%b: %b\r\n" % h
    ret += b"\r\n"
    return ret


def parse_credentials(
    header: str | None,
    prefixes: list | tuple | set | None = None,
) -> tuple[str | None, str | None]:
    """Parses any header with the aim to retrieve any credentials from it.

    Args:
        header (Optional[str]): The header to parse.
        prefixes (Optional[Union[List, Tuple, Set]], optional): The prefixes to look for. Defaults to None.

    Returns:
        Tuple[Optional[str], Optional[str]]: The prefix and the credentials.
    """  # noqa: E501
    if not prefixes or not isinstance(prefixes, (list, tuple, set)):
        prefixes = ("Basic", "Bearer", "Token")
    if header is not None:
        for prefix in prefixes:
            if prefix in header:
                return prefix, header.partition(prefix)[-1].strip()
    return None, header



import shutil

from argparse import ArgumentParser
from asyncio import run
from inspect import signature
from typing import Callable

from sanic import Sanic
from sanic.application.logo import get_logo
from sanic.cli.base import (
    SanicArgumentParser,
    SanicHelpFormatter,
)


def make_executor_parser(parser: ArgumentParser) -> None:
    parser.add_argument(
        "command",
        help="Command to execute",
    )


class ExecutorSubParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.description:
            self.description = ""
        self.description = get_logo(True) + self.description


class Executor:
    def __init__(self, app: Sanic, kwargs: dict) -> None:
        self.app = app
        self.kwargs = kwargs
        self.commands = self._make_commands()
        self.parser = self._make_parser()

    def run(self, command: str, args: list[str]) -> None:
        if command == "exec":
            args = ["--help"]
        parsed_args = self.parser.parse_args(args)
        if command not in self.commands:
            raise ValueError(f"Unknown command: {command}")
        parsed_kwargs = vars(parsed_args)
        parsed_kwargs.pop("command")
        run(self.commands[command](**parsed_kwargs))

    def _make_commands(self) -> dict[str, Callable]:
        commands = {c.name: c.func for c in self.app._future_commands}
        return commands

    def _make_parser(self) -> SanicArgumentParser:
        width = shutil.get_terminal_size().columns
        parser = SanicArgumentParser(
            prog="sanic",
            description=get_logo(True),
            formatter_class=lambda prog: SanicHelpFormatter(
                prog,
                max_help_position=36 if width > 96 else 24,
                indent_increment=4,
                width=None,
            ),
        )

        subparsers = parser.add_subparsers(
            dest="command",
            title="  Commands",
            parser_class=ExecutorSubParser,
        )
        for command in self.app._future_commands:
            sub = subparsers.add_parser(
                command.name,
                help=command.func.__doc__ or f"Execute {command.name}",
                formatter_class=SanicHelpFormatter,
            )
            self._add_arguments(sub, command.func)

        return parser

    def _add_arguments(self, parser: ArgumentParser, func: Callable) -> None:
        sig = signature(func)
        for param in sig.parameters.values():
            kwargs = {}
            if param.default is not param.empty:
                kwargs["default"] = param.default
            # In Python 3.14+, argparse validates help strings and rejects
            # non-string types. Convert annotations to string representation.
            help_text = None
            if param.annotation is not param.empty:
                if isinstance(param.annotation, str):
                    help_text = param.annotation
                else:
                    help_text = getattr(
                        param.annotation, "__name__", str(param.annotation)
                    )
            parser.add_argument(
                f"--{param.name}",
                help=help_text,
                **kwargs,
            )



"""
Sanic `provides a pattern
<https://sanicframework.org/guide/best-practices/exceptions.html#using-sanic-exceptions>`_
for providing a response when an exception occurs. However, if you do no handle
an exception, it will provide a fallback. There are three fallback types:

- HTML - *default*
- Text
- JSON

Setting ``app.config.FALLBACK_ERROR_FORMAT = "auto"`` will enable a switch that
will attempt to provide an appropriate response format based upon the
request type.
"""

from __future__ import annotations

import sys
import typing as t

from functools import partial
from traceback import extract_tb

from sanic.exceptions import BadRequest, SanicException
from sanic.helpers import STATUS_CODES
from sanic.log import deprecation, logger
from sanic.pages.error import ErrorPage
from sanic.response import html, json, text


dumps: t.Callable[..., str]
try:
    from ujson import dumps

    dumps = partial(dumps, escape_forward_slashes=False)
except ImportError:  # noqa
    from json import dumps

if t.TYPE_CHECKING:
    from sanic import HTTPResponse, Request

DEFAULT_FORMAT = "auto"
FALLBACK_TEXT = """\
The application encountered an unexpected error and could not continue.\
"""
FALLBACK_STATUS = 500
JSON = "application/json"


class BaseRenderer:
    """Base class that all renderers must inherit from.

    This class defines the structure for rendering objects, handling the core functionality that specific renderers may extend.

    Attributes:
        request (Request): The incoming request object that needs rendering.
        exception (Exception): Any exception that occurred and needs to be rendered.
        debug (bool): Flag indicating whether to render with debugging information.

    Methods:
        dumps: A static method that must be overridden by subclasses to define the specific rendering.

    Args:
        request (Request): The incoming request object that needs rendering.
        exception (Exception): Any exception that occurred and needs to be rendered.
        debug (bool): Flag indicating whether to render with debugging information.
    """  # noqa: E501

    dumps = staticmethod(dumps)

    def __init__(self, request: Request, exception: Exception, debug: bool):
        self.request = request
        self.exception = exception
        self.debug = debug

    @property
    def headers(self) -> t.Dict[str, str]:
        """The headers to be used for the response."""
        if isinstance(self.exception, SanicException):
            return getattr(self.exception, "headers", {})
        return {}

    @property
    def status(self):
        """The status code to be used for the response."""
        if isinstance(self.exception, SanicException):
            return getattr(self.exception, "status_code", FALLBACK_STATUS)
        return FALLBACK_STATUS

    @property
    def text(self):
        """The text to be used for the response."""
        if self.debug or isinstance(self.exception, SanicException):
            return str(self.exception)
        return FALLBACK_TEXT

    @property
    def title(self):
        """The title to be used for the response."""
        status_text = STATUS_CODES.get(self.status, b"Error Occurred").decode()
        return f"{self.status} — {status_text}"

    def render(self) -> HTTPResponse:
        """Outputs the exception as a response.

        Returns:
            HTTPResponse: The response object.
        """
        output = (
            self.full
            if self.debug and not getattr(self.exception, "quiet", False)
            else self.minimal
        )()
        output.status = self.status
        output.headers.update(self.headers)
        return output

    def minimal(self) -> HTTPResponse:  # noqa
        """Provide a formatted message that is meant to not show any sensitive data or details.

        This is the default fallback for production environments.

        Returns:
            HTTPResponse: The response object.
        """  # noqa: E501
        raise NotImplementedError

    def full(self) -> HTTPResponse:  # noqa
        """Provide a formatted message that has all details and is mean to be used primarily for debugging and non-production environments.

        Returns:
            HTTPResponse: The response object.
        """  # noqa: E501
        raise NotImplementedError


class HTMLRenderer(BaseRenderer):
    """Render an exception as HTML.

    The default fallback type.
    """

    def full(self) -> HTTPResponse:
        page = ErrorPage(
            debug=self.debug,
            title=super().title,
            text=super().text,
            request=self.request,
            exc=self.exception,
        )
        return html(page.render())

    def minimal(self) -> HTTPResponse:
        return self.full()


class TextRenderer(BaseRenderer):
    """Render an exception as plain text."""

    OUTPUT_TEXT = "{title}\n{bar}\n{text}\n\n{body}"
    SPACER = "  "

    def full(self) -> HTTPResponse:
        return text(
            self.OUTPUT_TEXT.format(
                title=self.title,
                text=self.text,
                bar=("=" * len(self.title)),
                body=self._generate_body(full=True),
            )
        )

    def minimal(self) -> HTTPResponse:
        return text(
            self.OUTPUT_TEXT.format(
                title=self.title,
                text=self.text,
                bar=("=" * len(self.title)),
                body=self._generate_body(full=False),
            )
        )

    @property
    def title(self):
        return f"⚠️ {super().title}"

    def _generate_body(self, *, full):
        lines = []
        if full:
            _, exc_value, __ = sys.exc_info()
            exceptions = []

            lines += [
                f"{self.exception.__class__.__name__}: {self.exception} while "
                f"handling path {self.request.path}",
                f"Traceback of {self.request.app.name} "
                "(most recent call last):\n",
            ]

            while exc_value:
                exceptions.append(self._format_exc(exc_value))
                exc_value = exc_value.__cause__

            lines += exceptions[::-1]

        for attr, display in (("context", True), ("extra", bool(full))):
            info = getattr(self.exception, attr, None)
            if info and display:
                lines += self._generate_object_display_list(info, attr)

        return "\n".join(lines)

    def _format_exc(self, exc):
        frames = "\n\n".join(
            [
                f"{self.SPACER * 2}File {frame.filename}, "
                f"line {frame.lineno}, in "
                f"{frame.name}\n{self.SPACER * 2}{frame.line}"
                for frame in extract_tb(exc.__traceback__)
            ]
        )
        return f"{self.SPACER}{exc.__class__.__name__}: {exc}\n{frames}"

    def _generate_object_display_list(self, obj, descriptor):
        lines = [f"\n{descriptor.title()}"]
        for key, value in obj.items():
            display = self.dumps(value)
            lines.append(f"{self.SPACER * 2}{key}: {display}")
        return lines


class JSONRenderer(BaseRenderer):
    """Render an exception as JSON."""

    def full(self) -> HTTPResponse:
        output = self._generate_output(full=True)
        return json(output, dumps=self.dumps)

    def minimal(self) -> HTTPResponse:
        output = self._generate_output(full=False)
        return json(output, dumps=self.dumps)

    def _generate_output(self, *, full):
        output = {
            "description": self.title,
            "status": self.status,
            "message": self.text,
        }

        for attr, display in (("context", True), ("extra", bool(full))):
            info = getattr(self.exception, attr, None)
            if info and display:
                output[attr] = info

        if full:
            _, exc_value, __ = sys.exc_info()
            exceptions = []

            while exc_value:
                exceptions.append(
                    {
                        "type": exc_value.__class__.__name__,
                        "exception": str(exc_value),
                        "frames": [
                            {
                                "file": frame.filename,
                                "line": frame.lineno,
                                "name": frame.name,
                                "src": frame.line,
                            }
                            for frame in extract_tb(exc_value.__traceback__)
                        ],
                    }
                )
                exc_value = exc_value.__cause__

            output["path"] = self.request.path
            output["args"] = self.request.args
            output["exceptions"] = exceptions[::-1]

        return output

    @property
    def title(self):
        return STATUS_CODES.get(self.status, b"Error Occurred").decode()


def escape(text):
    """Minimal HTML escaping, not for attribute values (unlike html.escape)."""
    return f"{text}".replace("&", "&amp;").replace("<", "&lt;")


MIME_BY_CONFIG = {
    "text": "text/plain",
    "json": "application/json",
    "html": "text/html",
}
CONFIG_BY_MIME = {v: k for k, v in MIME_BY_CONFIG.items()}
RENDERERS_BY_CONTENT_TYPE = {
    "text/plain": TextRenderer,
    "application/json": JSONRenderer,
    "multipart/form-data": HTMLRenderer,
    "text/html": HTMLRenderer,
}

# Handler source code is checked for which response types it returns with the
# route error_format="auto" (default) to determine which format to use.
RESPONSE_MAPPING = {
    "json": "json",
    "text": "text",
    "html": "html",
    "JSONResponse": "json",
    "text/plain": "text",
    "text/html": "html",
    "application/json": "json",
}


def check_error_format(format):
    """Check that the format is known."""
    if format not in MIME_BY_CONFIG and format != "auto":
        raise SanicException(f"Unknown format: {format}")


def exception_response(
    request: Request,
    exception: Exception,
    debug: bool,
    fallback: str,
    base: t.Type[BaseRenderer],
    renderer: t.Optional[t.Type[BaseRenderer]] = None,
) -> HTTPResponse:
    """Render a response for the default FALLBACK exception handler."""
    if not renderer:
        mt = guess_mime(request, fallback)
        renderer = RENDERERS_BY_CONTENT_TYPE.get(mt, base)

    renderer = t.cast(t.Type[BaseRenderer], renderer)
    return renderer(request, exception, debug).render()


def guess_mime(req: Request, fallback: str) -> str:
    """Guess the MIME type for the response based upon the request."""
    # Attempt to find a suitable MIME format for the response.
    # Insertion-ordered map of formats["html"] = "source of that suggestion"
    formats = {}
    name = ""
    # Route error_format (by magic from handler code if auto, the default)
    if req.route:
        name = req.route.name
        f = req.route.extra.error_format
        if f in MIME_BY_CONFIG:
            formats[f] = name

    if not formats and fallback in MIME_BY_CONFIG:
        formats[fallback] = "FALLBACK_ERROR_FORMAT"

    # If still not known, check for the request for clues of JSON
    if not formats and fallback == "auto" and req.accept.match(JSON):
        if JSON in req.accept:  # Literally, not wildcard
            formats["json"] = "request.accept"
        elif JSON in req.headers.getone("content-type", ""):
            formats["json"] = "content-type"
        # DEPRECATION: Remove this block in 24.3
        else:
            c = None
            try:
                c = req.json
            except BadRequest:
                pass
            if c:
                formats["json"] = "request.json"
                deprecation(
                    "Response type was determined by the JSON content of "
                    "the request. This behavior is deprecated and will be "
                    "removed in v24.3. Please specify the format either by\n"
                    f'  error_format="json" on route {name}, by\n'
                    '  FALLBACK_ERROR_FORMAT = "json", or by adding header\n'
                    "  accept: application/json to your requests.",
                    24.3,
                )

    # Any other supported formats
    if fallback == "auto":
        for k in MIME_BY_CONFIG:
            if k not in formats:
                formats[k] = "any"

    mimes = [MIME_BY_CONFIG[k] for k in formats]
    m = req.accept.match(*mimes)
    if m:
        format = CONFIG_BY_MIME[m.mime]
        source = formats[format]
        logger.debug(
            "Error Page: The client accepts %s, using '%s' from %s",
            m.header,
            format,
            source,
        )
    else:
        logger.debug(
            "Error Page: No format found, the client accepts %s",
            repr(req.accept),
        )
    return m.mime



from __future__ import annotations

from collections.abc import Mapping
from datetime import datetime
from inspect import isawaitable
from multiprocessing.connection import Connection
from os import environ
from pathlib import Path
from typing import Any

from sanic.exceptions import Unauthorized
from sanic.helpers import Default
from sanic.log import logger
from sanic.request import Request
from sanic.response import json


class Inspector:
    """Inspector for Sanic workers.

    This class is used to create an inspector for Sanic workers. It is
    instantiated by the worker class and is used to create a Sanic app
    that can be used to inspect and control the workers and the server.

    It is not intended to be used directly by the user.

    See [Inspector](/en/guide/deployment/inspector) for more information.

    Args:
        publisher (Connection): The connection to the worker.
        app_info (Dict[str, Any]): Information about the app.
        worker_state (Mapping[str, Any]): The state of the worker.
        host (str): The host to bind the inspector to.
        port (int): The port to bind the inspector to.
        api_key (str): The API key to use for authentication.
        tls_key (Union[Path, str, Default]): The path to the TLS key file.
        tls_cert (Union[Path, str, Default]): The path to the TLS cert file.
    """

    def __init__(
        self,
        publisher: Connection,
        app_info: dict[str, Any],
        worker_state: Mapping[str, Any],
        host: str,
        port: int,
        api_key: str,
        tls_key: Path | str | Default,
        tls_cert: Path | str | Default,
    ):
        self._publisher = publisher
        self.app_info = app_info
        self.worker_state = worker_state
        self.host = host
        self.port = port
        self.api_key = api_key
        self.tls_key = tls_key
        self.tls_cert = tls_cert

    def __call__(self, run=True, **_) -> Inspector:
        from sanic import Sanic

        self.app = Sanic("Inspector")
        self._setup()
        if run:
            self.app.run(
                host=self.host,
                port=self.port,
                single_process=True,
                ssl={"key": self.tls_key, "cert": self.tls_cert}
                if not isinstance(self.tls_key, Default)
                and not isinstance(self.tls_cert, Default)
                else None,
            )
        return self

    def _setup(self):
        self.app.get("/")(self._info)
        self.app.post("/<action:str>")(self._action)
        if self.api_key:
            self.app.on_request(self._authentication)
        environ["SANIC_IGNORE_PRODUCTION_WARNING"] = "true"

    def _authentication(self, request: Request) -> None:
        if request.token != self.api_key:
            raise Unauthorized("Bad API key")

    async def _action(self, request: Request, action: str):
        logger.info("Incoming inspector action: %s", action)
        output: Any = None
        method = getattr(self, action, None)
        if method:
            kwargs = {}
            if request.body:
                kwargs = request.json
            args = kwargs.pop("args", ())
            output = method(*args, **kwargs)
            if isawaitable(output):
                output = await output

        return await self._respond(request, output)

    async def _info(self, request: Request):
        return await self._respond(request, self._state_to_json())

    async def _respond(self, request: Request, output: Any):
        name = request.match_info.get("action", "info")
        return json({"meta": {"action": name}, "result": output})

    def _state_to_json(self) -> dict[str, Any]:
        output = {"info": self.app_info}
        output["workers"] = self._make_safe(dict(self.worker_state))
        return output

    @staticmethod
    def _make_safe(obj: dict[str, Any]) -> dict[str, Any]:
        for key, value in obj.items():
            if isinstance(value, dict):
                obj[key] = Inspector._make_safe(value)
            elif isinstance(value, datetime):
                obj[key] = value.isoformat()
        return obj

    def reload(self, zero_downtime: bool = False) -> None:
        """Reload the workers

        Args:
            zero_downtime (bool, optional): Whether to use zero downtime
                reload. Defaults to `False`.
        """
        message = "__ALL_PROCESSES__:"
        if zero_downtime:
            message += ":STARTUP_FIRST"
        self._publisher.send(message)

    def scale(self, replicas: str | int) -> str:
        """Scale the number of workers

        Args:
            replicas (Union[str, int]): The number of workers to scale to.

        Returns:
            str: A log message.
        """
        num_workers = 1
        if replicas:
            num_workers = int(replicas)
        log_msg = f"Scaling to {num_workers}"
        logger.info(log_msg)
        message = f"__SCALE__:{num_workers}"
        self._publisher.send(message)
        return log_msg

    def shutdown(self) -> None:
        """Shutdown the workers"""
        message = "__TERMINATE__"
        self._publisher.send(message)



from enum import IntEnum, auto

from sanic.compat import UpperStrEnum


class RestartOrder(UpperStrEnum):
    """Available restart orders."""

    SHUTDOWN_FIRST = auto()
    STARTUP_FIRST = auto()


class ProcessState(IntEnum):
    """Process states."""

    NONE = auto()
    IDLE = auto()
    RESTARTING = auto()
    STARTING = auto()
    STARTED = auto()
    ACKED = auto()
    JOINED = auto()
    TERMINATED = auto()
    FAILED = auto()
    COMPLETED = auto()



from typing import Any, NamedTuple


class REPLLocal(NamedTuple):
    var: Any
    name: str
    desc: str


class REPLContext:
    BUILTINS = {
        "app": "The Sanic application instance",
        "sanic": "The Sanic module",
        "do": "An async function to fake a request to the application",
        "client": "A client to access the Sanic app instance using httpx",
    }

    def __init__(self):
        self._locals: set[REPLLocal] = set()

    def add(
        self,
        var: Any,
        name: str | None = None,
        desc: str | None = None,
    ):
        """Add a local variable to be available in REPL context.

        Args:
            var (Any): A module, class, object or a class.
            name (Optional[str], optional): An alias for the local. Defaults to None.
            desc (Optional[str], optional): A brief description for the local. Defaults to None.
        """  # noqa: E501
        if name is None:
            try:
                name = var.__name__
            except AttributeError:
                name = var.__class__.__name__

        if desc is None:
            try:
                desc = var.__doc__ or ""
            except AttributeError:
                desc = str(type(var))

        assert isinstance(desc, str) and isinstance(
            name, str
        )  # Just to make mypy happy

        if name in self.BUILTINS:
            raise ValueError(f"Cannot override built-in variable: {name}")

        desc = self._truncate(desc)

        self._locals.add(REPLLocal(var, name, desc))

    def __setattr__(self, name: str, value: Any):
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            self.add(value, name)

    def __iter__(self):
        return iter(self._locals)

    @staticmethod
    def _truncate(s: str, limit: int = 40) -> str:
        s = s.replace("\n", " ")
        return s[:limit] + "..." if len(s) > limit else s



from __future__ import annotations

from sanic.log import error_logger
from sanic.worker.constants import RestartOrder
from sanic.worker.process import ProcessState, WorkerProcess


class Restarter:
    def restart(
        self,
        transient_processes: list[WorkerProcess],
        durable_processes: list[WorkerProcess],
        process_names: list[str] | None = None,
        restart_order=RestartOrder.SHUTDOWN_FIRST,
        **kwargs,
    ) -> None:
        """Restart the worker processes.

        Args:
            process_names (Optional[List[str]], optional): The names of the processes to restart.
                If `None`, then all processes will be restarted. Defaults to `None`.
            restart_order (RestartOrder, optional): The order in which to restart the processes.
                Defaults to `RestartOrder.SHUTDOWN_FIRST`.
        """  # noqa: E501
        restarted = self._restart_transient(
            transient_processes,
            process_names or [],
            restart_order,
            **kwargs,
        )
        restarted |= self._restart_durable(
            durable_processes,
            process_names or [],
            restart_order,
            **kwargs,
        )

        if process_names and not restarted:
            error_logger.error(
                f"Failed to restart processes: {', '.join(process_names)}"
            )

    def _restart_transient(
        self,
        processes: list[WorkerProcess],
        process_names: list[str],
        restart_order: RestartOrder,
        **kwargs,
    ) -> set[str]:
        restarted: set[str] = set()
        for process in processes:
            if not process.restartable or (
                process_names and process.name not in process_names
            ):
                continue
            self._restart_process(process, restart_order, **kwargs)
            restarted.add(process.name)
        return restarted

    def _restart_durable(
        self,
        processes: list[WorkerProcess],
        process_names: list[str],
        restart_order: RestartOrder,
        **kwargs,
    ) -> set[str]:
        restarted: set[str] = set()
        if not process_names:
            return restarted
        for process in processes:
            if not process.restartable or process.name not in process_names:
                continue
            if process.state not in (
                ProcessState.COMPLETED,
                ProcessState.FAILED,
                ProcessState.NONE,
            ):
                error_logger.error(
                    f"Cannot restart process {process.name} because "
                    "it is not in a final state. Current state is: "
                    f"{process.state.name}."
                )
                continue
            self._restart_process(process, restart_order, **kwargs)
            restarted.add(process.name)

        return restarted

    def _restart_process(self, process, restart_order, **kwargs):
        process.restart(restart_order=restart_order, **kwargs)



from __future__ import annotations

import os
import platform

from asyncio import (
    AbstractEventLoop,
    CancelledError,
    Protocol,
    all_tasks,
    get_event_loop,
    get_running_loop,
    new_event_loop,
)
from collections.abc import Mapping
from contextlib import suppress
from functools import partial
from importlib import import_module
from multiprocessing import (
    Manager,
    Pipe,
    get_context,
    get_start_method,
    set_start_method,
)
from multiprocessing.context import BaseContext
from pathlib import Path
from socket import SHUT_RDWR, socket
from ssl import SSLContext
from time import sleep
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Literal,
    cast,
)

from sanic.application.ext import setup_ext
from sanic.application.logo import get_logo
from sanic.application.motd import MOTD
from sanic.application.state import ApplicationServerInfo, Mode, ServerStage
from sanic.base.meta import SanicMeta
from sanic.compat import OS_IS_WINDOWS, StartMethod
from sanic.exceptions import ServerError, ServerKilled
from sanic.helpers import Default, _default, is_atty
from sanic.http.constants import HTTP
from sanic.http.tls import get_ssl_context, process_to_context
from sanic.http.tls.context import SanicSSLContext
from sanic.log import Colors, deprecation, error_logger, logger
from sanic.logging.setup import setup_logging
from sanic.models.handler_types import ListenerType
from sanic.server import Signal as ServerSignal
from sanic.server import try_use_uvloop
from sanic.server.async_server import AsyncioServer
from sanic.server.events import trigger_events
from sanic.server.goodbye import get_goodbye
from sanic.server.loop import try_windows_loop
from sanic.server.protocols.http_protocol import HttpProtocol
from sanic.server.protocols.websocket_protocol import WebSocketProtocol
from sanic.server.runners import serve
from sanic.server.socket import configure_socket, remove_unix_socket
from sanic.worker.loader import AppLoader
from sanic.worker.manager import WorkerManager
from sanic.worker.multiplexer import WorkerMultiplexer
from sanic.worker.reloader import Reloader
from sanic.worker.serve import worker_serve


if TYPE_CHECKING:
    from sanic import Sanic
    from sanic.application.state import ApplicationState
    from sanic.config import Config

SANIC_PACKAGES = ("sanic-routing", "sanic-testing", "sanic-ext")


HTTPVersion = HTTP | Literal[1] | Literal[3]


class StartupMixin(metaclass=SanicMeta):
    _app_registry: ClassVar[dict[str, Sanic]]
    name: str
    asgi: bool
    config: Config
    listeners: dict[str, list[ListenerType[Any]]]
    state: ApplicationState
    websocket_enabled: bool
    multiplexer: WorkerMultiplexer

    test_mode: ClassVar[bool]
    start_method: ClassVar[StartMethod] = _default
    START_METHOD_SET: ClassVar[bool] = False

    def setup_loop(self) -> None:
        """Set up the event loop.

        An internal method that sets up the event loop to uvloop if
        possible, or a Windows selector loop if on Windows.

        Returns:
            None
        """
        if not self.asgi:
            if self.config.USE_UVLOOP is True or (
                isinstance(self.config.USE_UVLOOP, Default)
                and not OS_IS_WINDOWS
            ):
                try_use_uvloop()
            elif OS_IS_WINDOWS:
                try_windows_loop()

    @property
    def m(self) -> WorkerMultiplexer:
        """Interface for interacting with the worker processes

        This is a shortcut for `app.multiplexer`. It is available only in a
        worker process using the Sanic server. It allows you to interact with
        the worker processes, such as sending messages and commands.

        See [Access to the multiplexer](/en/guide/deployment/manager#access-to-the-multiplexer) for more information.

        Returns:
            WorkerMultiplexer: The worker multiplexer instance

        Examples:
            ```python
            app.m.restart()    # restarts the worker
            app.m.terminate()  # terminates the worker
            app.m.scale(4)     # scales the number of workers to 4
        ```
        """  # noqa: E501
        return self.multiplexer

    def make_coffee(self, *args, **kwargs):
        """
        Try for yourself! `sanic server:app --coffee`

         ```
         ▄████████▄
        ██       ██▀▀▄
        ███████████  █
        ███████████▄▄▀
         ▀███████▀

         ```
        """
        self.state.coffee = True
        self.run(*args, **kwargs)

    def run(
        self,
        host: str | None = None,
        port: int | None = None,
        *,
        dev: bool = False,
        debug: bool = False,
        auto_reload: bool | None = None,
        version: HTTPVersion = HTTP.VERSION_1,
        ssl: None | SSLContext | dict | str | list | tuple = None,
        sock: socket | None = None,
        workers: int = 1,
        protocol: type[Protocol] | None = None,
        backlog: int = 100,
        register_sys_signals: bool = True,
        access_log: bool | None = None,
        unix: str | None = None,
        loop: AbstractEventLoop | None = None,
        reload_dir: list[str] | str | None = None,
        noisy_exceptions: bool | None = None,
        motd: bool = True,
        fast: bool = False,
        verbosity: int = 0,
        motd_display: dict[str, str] | None = None,
        auto_tls: bool = False,
        single_process: bool = False,
    ) -> None:
        """Run the HTTP Server and listen until keyboard interrupt or term signal. On termination, drain connections before closing.

        .. note::
            When you need control over running the Sanic instance, this is the method to use.
            However, in most cases the preferred method is to use the CLI command:

            ```sh
            sanic server:app`
            ```

        If you are using this method to run Sanic, make sure you do the following:

        1. Use `if __name__ == "__main__"` to guard the code.
        2. Do **NOT** define the app instance inside the `if` block.

        See [Dynamic Applications](/en/guide/deployment/app-loader) for more information about the second point.

        Args:
            host (Optional[str]): Address to host on.
            port (Optional[int]): Port to host on.
            dev (bool): Run the server in development mode.
            debug (bool): Enables debug output (slows server).
            auto_reload (Optional[bool]): Reload app whenever its source code is changed.
                Enabled by default in debug mode.
            version (HTTPVersion): HTTP Version.
            ssl (Union[None, SSLContext, dict, str, list, tuple]): SSLContext, or location of certificate and key
                for SSL encryption of worker(s).
            sock (Optional[socket]): Socket for the server to accept connections from.
            workers (int): Number of processes received before it is respected.
            protocol (Optional[Type[Protocol]]): Subclass of asyncio Protocol class.
            backlog (int): A number of unaccepted connections that the system will allow
                before refusing new connections.
            register_sys_signals (bool): Register SIG* events.
            access_log (Optional[bool]): Enables writing access logs (slows server).
            unix (Optional[str]): Unix socket to listen on instead of TCP port.
            loop (Optional[AbstractEventLoop]): AsyncIO event loop.
            reload_dir (Optional[Union[List[str], str]]): Directory to watch for code changes, if auto_reload is True.
            noisy_exceptions (Optional[bool]): Log exceptions that are normally considered to be quiet/silent.
            motd (bool): Display Message of the Day.
            fast (bool): Enable fast mode.
            verbosity (int): Verbosity level.
            motd_display (Optional[Dict[str, str]]): Customize Message of the Day display.
            auto_tls (bool): Enable automatic TLS certificate handling.
            single_process (bool): Enable single process mode.

        Returns:
            None

        Raises:
            RuntimeError: Raised when attempting to serve HTTP/3 as a secondary server.
            RuntimeError: Raised when attempting to use both `fast` and `workers`.
            RuntimeError: Raised when attempting to use `single_process` with `fast`, `workers`, or `auto_reload`.
            TypeError: Raised when attempting to use `loop` with `create_server`.
            ValueError: Raised when `PROXIES_COUNT` is negative.

        Examples:
            ```python
            from sanic import Sanic, Request, json

            app = Sanic("TestApp")


            @app.get("/")
            async def handler(request: Request):
                return json({"foo": "bar"})


            if __name__ == "__main__":
                app.run(port=9999, dev=True)
            ```
        """  # noqa: E501
        self.prepare(
            host=host,
            port=port,
            dev=dev,
            debug=debug,
            auto_reload=auto_reload,
            version=version,
            ssl=ssl,
            sock=sock,
            workers=workers,
            protocol=protocol,
            backlog=backlog,
            register_sys_signals=register_sys_signals,
            access_log=access_log,
            unix=unix,
            loop=loop,
            reload_dir=reload_dir,
            noisy_exceptions=noisy_exceptions,
            motd=motd,
            fast=fast,
            verbosity=verbosity,
            motd_display=motd_display,
            auto_tls=auto_tls,
            single_process=single_process,
        )

        if single_process:
            serve = self.__class__.serve_single
        else:
            serve = self.__class__.serve
        serve(primary=self)  # type: ignore

    def prepare(
        self,
        host: str | None = None,
        port: int | None = None,
        *,
        dev: bool = False,
        debug: bool = False,
        auto_reload: bool | None = None,
        version: HTTPVersion = HTTP.VERSION_1,
        ssl: None | SSLContext | dict | str | list | tuple = None,
        sock: socket | None = None,
        workers: int = 1,
        protocol: type[Protocol] | None = None,
        backlog: int = 100,
        register_sys_signals: bool = True,
        access_log: bool | None = None,
        unix: str | None = None,
        loop: AbstractEventLoop | None = None,
        reload_dir: list[str] | str | None = None,
        noisy_exceptions: bool | None = None,
        motd: bool = True,
        fast: bool = False,
        verbosity: int = 0,
        motd_display: dict[str, str] | None = None,
        coffee: bool = False,
        auto_tls: bool = False,
        single_process: bool = False,
    ) -> None:
        """Prepares one or more Sanic applications to be served simultaneously.

        This low-level API is typically used when you need to run multiple Sanic applications at the same time. Once prepared, `Sanic.serve()` should be called in the `if __name__ == "__main__"` block.

        .. note::
            "Preparing" and "serving" with this function is equivalent to using `app.run` for a single instance. This should only be used when running multiple applications at the same time.

        Args:
            host (Optional[str], optional): Hostname to listen on. Defaults to `None`.
            port (Optional[int], optional): Port to listen on. Defaults to `None`.
            dev (bool, optional): Development mode. Defaults to `False`.
            debug (bool, optional): Debug mode. Defaults to `False`.
            auto_reload (Optional[bool], optional): Auto reload feature. Defaults to `None`.
            version (HTTPVersion, optional): HTTP version to use. Defaults to `HTTP.VERSION_1`.
            ssl (Union[None, SSLContext, dict, str, list, tuple], optional): SSL configuration. Defaults to `None`.
            sock (Optional[socket], optional): Socket to bind to. Defaults to `None`.
            workers (int, optional): Number of worker processes. Defaults to `1`.
            protocol (Optional[Type[Protocol]], optional): Custom protocol class. Defaults to `None`.
            backlog (int, optional): Maximum number of pending connections. Defaults to `100`.
            register_sys_signals (bool, optional): Register system signals. Defaults to `True`.
            access_log (Optional[bool], optional): Access log. Defaults to `None`.
            unix (Optional[str], optional): Unix socket. Defaults to `None`.
            loop (Optional[AbstractEventLoop], optional): Event loop. Defaults to `None`.
            reload_dir (Optional[Union[List[str], str]], optional): Reload directory. Defaults to `None`.
            noisy_exceptions (Optional[bool], optional): Display exceptions. Defaults to `None`.
            motd (bool, optional): Display message of the day. Defaults to `True`.
            fast (bool, optional): Fast mode. Defaults to `False`.
            verbosity (int, optional): Verbosity level. Defaults to `0`.
            motd_display (Optional[Dict[str, str]], optional): Custom MOTD display. Defaults to `None`.
            coffee (bool, optional): Coffee mode. Defaults to `False`.
            auto_tls (bool, optional): Auto TLS. Defaults to `False`.
            single_process (bool, optional): Single process mode. Defaults to `False`.

        Raises:
            RuntimeError: Raised when attempting to serve HTTP/3 as a secondary server.
            RuntimeError: Raised when attempting to use both `fast` and `workers`.
            RuntimeError: Raised when attempting to use `single_process` with `fast`, `workers`, or `auto_reload`.
            TypeError: Raised when attempting to use `loop` with `create_server`.
            ValueError: Raised when `PROXIES_COUNT` is negative.

        Examples:
            ```python
            if __name__ == "__main__":
                app.prepare()
                app.serve()
            ```
        """  # noqa: E501
        if version == 3 and self.state.server_info:
            raise RuntimeError(
                "Serving HTTP/3 instances as a secondary server is "
                "not supported. There can only be a single HTTP/3 worker "
                "and it must be the first instance prepared."
            )

        if dev:
            debug = True
            auto_reload = True

        if debug and access_log is None:
            access_log = True

        self.state.verbosity = verbosity
        if not self.state.auto_reload:
            self.state.auto_reload = bool(auto_reload)

        if fast and workers != 1:
            raise RuntimeError("You cannot use both fast=True and workers=X")

        if single_process and (fast or (workers > 1) or auto_reload):
            raise RuntimeError(
                "Single process cannot be run with multiple workers "
                "or auto-reload"
            )

        if register_sys_signals is False and not single_process:
            raise RuntimeError(
                "Cannot run Sanic.serve with register_sys_signals=False. "
                "Use Sanic.serve_single."
            )

        if motd_display:
            self.config.MOTD_DISPLAY.update(motd_display)

        if reload_dir:
            if isinstance(reload_dir, str):
                reload_dir = [reload_dir]

            for directory in reload_dir:
                direc = Path(directory)
                if not direc.is_dir():
                    logger.warning(
                        f"Directory {directory} could not be located"
                    )
                self.state.reload_dirs.add(Path(directory))

        if loop is not None:
            raise TypeError(
                "loop is not a valid argument. To use an existing loop, "
                "change to create_server().\nSee more: "
                "https://sanic.readthedocs.io/en/latest/sanic/deploying.html"
                "#asynchronous-support"
            )

        if sock is None:
            host, port = self.get_address(host, port, version, auto_tls)

        if protocol is None:
            protocol = (
                WebSocketProtocol if self.websocket_enabled else HttpProtocol
            )

        # Set explicitly passed configuration values
        for attribute, value in {
            "ACCESS_LOG": access_log,
            "AUTO_RELOAD": auto_reload,
            "MOTD": motd,
            "NOISY_EXCEPTIONS": noisy_exceptions,
        }.items():
            if value is not None:
                setattr(self.config, attribute, value)

        if fast:
            self.state.fast = True
            try:
                workers = len(os.sched_getaffinity(0))
            except AttributeError:  # no cov
                workers = os.cpu_count() or 1

        if coffee:
            self.state.coffee = True

        server_settings = self._helper(
            host=host,
            port=port,
            debug=debug,
            version=version,
            ssl=ssl,
            sock=sock,
            unix=unix,
            workers=workers,
            protocol=protocol,
            backlog=backlog,
            register_sys_signals=register_sys_signals,
            auto_tls=auto_tls,
        )
        self.state.server_info.append(
            ApplicationServerInfo(settings=server_settings)
        )

        # if self.config.USE_UVLOOP is True or (
        #     self.config.USE_UVLOOP is _default and not OS_IS_WINDOWS
        # ):
        #     try_use_uvloop()

    async def create_server(
        self,
        host: str | None = None,
        port: int | None = None,
        *,
        debug: bool = False,
        ssl: None | SSLContext | dict | str | list | tuple = None,
        sock: socket | None = None,
        protocol: type[Protocol] | None = None,
        backlog: int = 100,
        access_log: bool | None = None,
        unix: str | None = None,
        return_asyncio_server: bool = True,
        asyncio_server_kwargs: dict[str, Any] | None = None,
        noisy_exceptions: bool | None = None,
    ) -> AsyncioServer | None:
        """
        Low level API for creating a Sanic Server instance.

        This method will create a Sanic Server instance, but will not start
        it. This is useful for integrating Sanic into other systems. But, you
        should take caution when using it as it is a low level API and does
        not perform any of the lifecycle events.

        .. note::
            This does not support multiprocessing and is not the preferred
            way to run a Sanic application. Proceed with caution.

        You will need to start the server yourself as shown in the example
        below. You are responsible for the lifecycle of the server, including
        app startup using `await app.startup()`. No events will be triggered
        for you, so you will need to trigger them yourself if wanted.

        Args:
            host (Optional[str]): Address to host on.
            port (Optional[int]): Port to host on.
            debug (bool): Enables debug output (slows server).
            ssl (Union[None, SSLContext, dict, str, list, tuple]): SSLContext,
                or location of certificate and key for SSL encryption
                of worker(s).
            sock (Optional[socket]): Socket for the server to accept
                connections from.
            protocol (Optional[Type[Protocol]]): Subclass of
                `asyncio.Protocol` class.
            backlog (int): Number of unaccepted connections that the system
                will allow before refusing new connections.
            access_log (Optional[bool]): Enables writing access logs
                (slows server).
            return_asyncio_server (bool): _DEPRECATED_
            asyncio_server_kwargs (Optional[Dict[str, Any]]): Key-value
                arguments for asyncio/uvloop `create_server` method.
            noisy_exceptions (Optional[bool]): Log exceptions that are normally
                considered to be quiet/silent.

        Returns:
            Optional[AsyncioServer]: AsyncioServer if `return_asyncio_server`
                is `True` else `None`.

        Examples:
            ```python
            import asyncio
            import uvloop
            from sanic import Sanic, response


            app = Sanic("Example")


            @app.route("/")
            async def test(request):
                return response.json({"answer": "42"})


            async def main():
                server = await app.create_server()
                await server.startup()
                await server.serve_forever()


            if __name__ == "__main__":
                asyncio.set_event_loop(uvloop.new_event_loop())
                asyncio.run(main())
            ```
        """

        if sock is None:
            host, port = host, port = self.get_address(host, port)

        if protocol is None:
            protocol = (
                WebSocketProtocol if self.websocket_enabled else HttpProtocol
            )

        # Set explicitly passed configuration values
        for attribute, value in {
            "ACCESS_LOG": access_log,
            "NOISY_EXCEPTIONS": noisy_exceptions,
        }.items():
            if value is not None:
                setattr(self.config, attribute, value)

        if not return_asyncio_server:
            return_asyncio_server = True
            deprecation(
                "The `return_asyncio_server` argument is deprecated and "
                "ignored. It will be removed in v24.3.",
                24.3,
            )

        server_settings = self._helper(
            host=host,
            port=port,
            debug=debug,
            ssl=ssl,
            sock=sock,
            unix=unix,
            loop=get_event_loop(),
            protocol=protocol,
            backlog=backlog,
            run_async=return_asyncio_server,
        )

        if not isinstance(self.config.USE_UVLOOP, Default):
            error_logger.warning(
                "You are trying to change the uvloop configuration, but "
                "this is only effective when using the run(...) method. "
                "When using the create_server(...) method Sanic will use "
                "the already existing loop."
            )

        main_start = server_settings.pop("main_start", None)
        main_stop = server_settings.pop("main_stop", None)
        if main_start or main_stop:
            logger.warning(
                "Listener events for the main process are not available "
                "with create_server()"
            )

        return await serve(
            asyncio_server_kwargs=asyncio_server_kwargs, **server_settings
        )

    def stop(self, terminate: bool = True, unregister: bool = False) -> None:
        """This kills the Sanic server, cleaning up after itself.

        Args:
            terminate (bool): Force kill all requests immediately without
                allowing them to finish processing.
            unregister (bool): Unregister the app from the global registry.

        Returns:
            None
        """
        if terminate and hasattr(self, "multiplexer"):
            self.multiplexer.terminate()
        if self.state.stage is not ServerStage.STOPPED:
            self.shutdown_tasks(timeout=0)  # type: ignore
            for task in all_tasks():
                with suppress(AttributeError):
                    if task.get_name() == "RunServer":
                        task.cancel()
            get_event_loop().stop()

        if unregister:
            self.__class__.unregister_app(self)  # type: ignore

    def _helper(
        self,
        host: str | None = None,
        port: int | None = None,
        debug: bool = False,
        version: HTTPVersion = HTTP.VERSION_1,
        ssl: None | SSLContext | dict | str | list | tuple = None,
        sock: socket | None = None,
        unix: str | None = None,
        workers: int = 1,
        loop: AbstractEventLoop | None = None,
        protocol: type[Protocol] = HttpProtocol,
        backlog: int = 100,
        register_sys_signals: bool = True,
        run_async: bool = False,
        auto_tls: bool = False,
    ) -> dict[str, Any]:
        """Helper function used by `run` and `create_server`."""
        if self.config.PROXIES_COUNT and self.config.PROXIES_COUNT < 0:
            raise ValueError(
                "PROXIES_COUNT cannot be negative. "
                "https://sanic.readthedocs.io/en/latest/sanic/config.html"
                "#proxy-configuration"
            )

        if not self.state.is_debug:
            self.state.mode = Mode.DEBUG if debug else Mode.PRODUCTION

        setup_logging(
            self.state.is_debug, self.config.NO_COLOR, self.config.LOG_EXTRA
        )

        if isinstance(version, int):
            version = HTTP(version)

        ssl = process_to_context(ssl)
        if version is HTTP.VERSION_3 or auto_tls:
            if TYPE_CHECKING:
                self = cast(Sanic, self)
            ssl = get_ssl_context(self, ssl)

        self.state.host = host or ""
        self.state.port = port or 0
        self.state.workers = workers
        self.state.ssl = ssl
        self.state.unix = unix
        self.state.sock = sock

        server_settings = {
            "protocol": protocol,
            "host": host,
            "port": port,
            "version": version,
            "sock": sock,
            "unix": unix,
            "ssl": ssl,
            "app": self,
            "signal": ServerSignal(),
            "loop": loop,
            "register_sys_signals": register_sys_signals,
            "backlog": backlog,
        }

        self.motd(server_settings=server_settings)

        if (
            is_atty()
            and not self.state.is_debug
            and not os.environ.get("SANIC_IGNORE_PRODUCTION_WARNING")
        ):
            error_logger.warning(
                f"{Colors.YELLOW}Sanic is running in PRODUCTION mode. "
                "Consider using '--debug' or '--dev' while actively "
                f"developing your application.{Colors.END}"
            )

        # Register start/stop events
        for event_name, settings_name, reverse in (
            ("main_process_start", "main_start", False),
            ("main_process_stop", "main_stop", True),
        ):
            listeners = self.listeners[event_name].copy()
            if reverse:
                listeners.reverse()
            # Prepend sanic to the arguments when listeners are triggered
            listeners = [partial(listener, self) for listener in listeners]
            server_settings[settings_name] = listeners  # type: ignore

        if run_async:
            server_settings["run_async"] = True

        return server_settings

    def motd(
        self,
        server_settings: dict[str, Any] | None = None,
    ) -> None:
        """Outputs the message of the day (MOTD).

        It generally can only be called once per process, and is usually
        called by the `run` method in the main process.

        Args:
            server_settings (Optional[Dict[str, Any]], optional): Settings for
                the server. Defaults to `None`.

        Returns:
            None
        """
        if (
            os.environ.get("SANIC_WORKER_NAME")
            or os.environ.get("SANIC_MOTD_OUTPUT")
            or os.environ.get("SANIC_WORKER_PROCESS")
            or os.environ.get("SANIC_SERVER_RUNNING")
        ):
            return
        serve_location = self.get_server_location(server_settings)
        if self.config.MOTD:
            logo = get_logo(coffee=self.state.coffee)
            display, extra = self.get_motd_data(server_settings)

            MOTD.output(logo, serve_location, display, extra)

    def get_motd_data(
        self, server_settings: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Retrieves the message of the day (MOTD) data.

        Args:
            server_settings (Optional[Dict[str, Any]], optional): Settings for
                the server. Defaults to `None`.

        Returns:
            Tuple[Dict[str, Any], Dict[str, Any]]: A tuple containing two
                dictionaries with the relevant MOTD data.
        """

        mode = [f"{self.state.mode},"]
        if self.state.fast:
            mode.append("goin' fast")
        if self.state.asgi:
            mode.append("ASGI")
        else:
            if self.state.workers == 1:
                mode.append("single worker")
            else:
                mode.append(f"w/ {self.state.workers} workers")

        if server_settings:
            server = ", ".join(
                (
                    self.state.server,
                    server_settings["version"].display(),  # type: ignore
                )
            )
        else:
            server = "ASGI" if self.asgi else "unknown"  # type: ignore

        display = {
            "app": self.name,
            "mode": " ".join(mode),
            "server": server,
            "python": platform.python_version(),
            "platform": platform.platform(),
        }
        extra = {}
        if self.config.AUTO_RELOAD:
            reload_display = "enabled"
            if self.state.reload_dirs:
                reload_display += ", ".join(
                    [
                        "",
                        *(
                            str(path.absolute())
                            for path in self.state.reload_dirs
                        ),
                    ]
                )
            display["auto-reload"] = reload_display

        packages = []
        for package_name in SANIC_PACKAGES:
            module_name = package_name.replace("-", "_")
            try:
                module = import_module(module_name)
                packages.append(f"{package_name}=={module.__version__}")  # type: ignore
            except ImportError:  # no cov
                ...

        if packages:
            display["packages"] = ", ".join(packages)

        if self.config.MOTD_DISPLAY:
            extra.update(self.config.MOTD_DISPLAY)

        return display, extra

    @property
    def serve_location(self) -> str:
        """Retrieve the server location.

        Returns:
            str: The server location.
        """
        try:
            server_settings = self.state.server_info[0].settings
            return self.get_server_location(server_settings)
        except IndexError:
            location = "ASGI" if self.asgi else "unknown"  # type: ignore
            return f"http://<{location}>"

    @staticmethod
    def get_server_location(
        server_settings: dict[str, Any] | None = None,
    ) -> str:
        """Using the server settings, retrieve the server location.

        Args:
            server_settings (Optional[Dict[str, Any]], optional): Settings for
                the server. Defaults to `None`.

        Returns:
            str: The server location.
        """
        serve_location = ""
        proto = "http"
        if not server_settings:
            return serve_location

        host = server_settings["host"]
        port = server_settings["port"]

        if server_settings.get("ssl") is not None:
            proto = "https"
        if server_settings.get("unix"):
            serve_location = f"{server_settings['unix']} {proto}://..."
        elif server_settings.get("sock"):
            host, port, *_ = server_settings["sock"].getsockname()

        if not serve_location and host and port:
            # colon(:) is legal for a host only in an ipv6 address
            display_host = f"[{host}]" if ":" in host else host
            serve_location = f"{proto}://{display_host}:{port}"

        return serve_location

    @staticmethod
    def get_address(
        host: str | None,
        port: int | None,
        version: HTTPVersion = HTTP.VERSION_1,
        auto_tls: bool = False,
    ) -> tuple[str, int]:
        """Retrieve the host address and port, with default values based on the given parameters.

        Args:
            host (Optional[str]): Host IP or FQDN for the service to use. Defaults to `"127.0.0.1"`.
            port (Optional[int]): Port number. Defaults to `8443` if version is 3 or `auto_tls=True`, else `8000`
            version (HTTPVersion, optional): HTTP Version. Defaults to `HTTP.VERSION_1` (HTTP/1.1).
            auto_tls (bool, optional): Automatic TLS flag. Defaults to `False`.

        Returns:
            Tuple[str, int]: Tuple containing the host and port
        """  # noqa: E501
        host = host or "127.0.0.1"
        port = port or (8443 if (version == 3 or auto_tls) else 8000)
        return host, port

    @classmethod
    def should_auto_reload(cls) -> bool:
        """Check if any applications have auto-reload enabled.

        Returns:
            bool: `True` if any applications have auto-reload enabled, else
                `False`.
        """
        return any(app.state.auto_reload for app in cls._app_registry.values())

    @classmethod
    def _get_startup_method(cls) -> str:
        return (
            cls.start_method
            if not isinstance(cls.start_method, Default)
            else "spawn"
        )

    @classmethod
    def _set_startup_method(cls) -> None:
        if cls.START_METHOD_SET and not cls.test_mode:
            return

        method = cls._get_startup_method()
        try:
            set_start_method(method, force=cls.test_mode)
        except RuntimeError:
            ctx = get_context()
            actual = ctx.get_start_method()
            if actual != method:
                raise RuntimeError(
                    f"Start method '{method}' was requested, but '{actual}' "
                    "was already set.\nFor more information, see: "
                    "https://sanic.dev/en/guide/running/manager.html#overcoming-a-coderuntimeerrorcode"
                ) from None
            else:
                raise
        cls.START_METHOD_SET = True

    @classmethod
    def _get_context(cls) -> BaseContext:
        method = cls._get_startup_method()
        logger.debug("Creating multiprocessing context using '%s'", method)
        actual = get_start_method()
        if method != actual:
            raise RuntimeError(
                f"Start method '{method}' was requested, but '{actual}' "
                "was already set.\nFor more information, see: "
                "https://sanic.dev/en/guide/running/manager.html#overcoming-a-coderuntimeerrorcode"
            ) from None
        return get_context()

    @classmethod
    def serve(
        cls,
        primary: Sanic | None = None,
        *,
        app_loader: AppLoader | None = None,
        factory: Callable[[], Sanic] | None = None,
    ) -> None:
        """Serve one or more Sanic applications.

        This is the main entry point for running Sanic applications. It
        should be called in the `if __name__ == "__main__"` block.

        Args:
            primary (Optional[Sanic], optional): The primary Sanic application
                to serve. Defaults to `None`.
            app_loader (Optional[AppLoader], optional): An AppLoader instance
                to use for loading applications. Defaults to `None`.
            factory (Optional[Callable[[], Sanic]], optional): A factory
                function to use for loading applications. Defaults to `None`.

        Raises:
            RuntimeError: Raised when no applications are found.
            RuntimeError: Raised when no server information is found for the
                primary application.
            RuntimeError: Raised when attempting to use `loop` with
                `create_server`.
            RuntimeError: Raised when attempting to use `single_process` with
                `fast`, `workers`, or `auto_reload`.
            RuntimeError: Raised when attempting to serve HTTP/3 as a
                secondary server.
            RuntimeError: Raised when attempting to use both `fast` and
                `workers`.
            TypeError: Raised when attempting to use `loop` with
                `create_server`.
            ValueError: Raised when `PROXIES_COUNT` is negative.

        Examples:
            ```python
            if __name__ == "__main__":
                app.prepare()
                Sanic.serve()
            ```
        """
        cls._set_startup_method()
        os.environ["SANIC_MOTD_OUTPUT"] = "true"
        apps = list(cls._app_registry.values())
        if factory:
            primary = factory()
        else:
            if not primary:
                if app_loader:
                    primary = app_loader.load()
                if not primary:
                    try:
                        primary = apps[0]
                    except IndexError:
                        raise RuntimeError(
                            "Did not find any applications."
                        ) from None

            # This exists primarily for unit testing
            if not primary.state.server_info:  # no cov
                for app in apps:
                    app.state.server_info.clear()
                return

        try:
            primary_server_info = primary.state.server_info[0]
        except IndexError:
            raise RuntimeError(
                f"No server information found for {primary.name}. Perhaps you "
                "need to run app.prepare(...)?"
            ) from None

        socks = []
        try:
            sync_manager = Manager()
        except EOFError:
            message = (
                "Sanic server could not start: worker process failed.\n\n"
                "This may have happened if you are running Sanic in the "
                "global scope and not inside of a "
                '`if __name__ == "__main__"` block.\n\nSee more information: '
                "https://sanic.dev/en/guide/deployment/manager.html#"
                "how-sanic-server-starts-processes\n"
            )
            if not OS_IS_WINDOWS:
                message += (
                    "\nAlternatively, you can set "
                    '`Sanic.start_method = "fork"` at the start of your '
                    "application to avoid this issue (Unix only).\n"
                )
            raise ServerError(message, quiet=True) from None
        worker_state: Mapping[str, Any] = {"state": "NONE"}
        setup_ext(primary)
        exit_code = 0
        workers_started = False
        try:
            primary_server_info.settings.pop("main_start", None)
            primary_server_info.settings.pop("main_stop", None)
            main_start = primary.listeners.get("main_process_start")
            main_stop = primary.listeners.get("main_process_stop")
            app = primary_server_info.settings.pop("app")
            app.setup_loop()
            loop = new_event_loop()
            trigger_events(main_start, loop, primary)

            socks = [
                sock
                for sock in [
                    configure_socket(server_info.settings)
                    for app in apps
                    for server_info in app.state.server_info
                ]
                if sock
            ]
            primary_server_info.settings["run_multiple"] = True
            monitor_sub, monitor_pub = Pipe(True)
            worker_state = sync_manager.dict()
            kwargs: dict[str, Any] = {
                **primary_server_info.settings,
                "monitor_publisher": monitor_pub,
                "worker_state": worker_state,
            }

            if not app_loader:
                if factory:
                    app_loader = AppLoader(factory=factory)
                else:
                    app_loader = AppLoader(
                        factory=partial(cls.get_app, app.name)  # type: ignore
                    )
            kwargs["app_name"] = app.name
            kwargs["app_loader"] = app_loader
            kwargs["server_info"] = {}
            kwargs["passthru"] = {
                "auto_reload": app.auto_reload,
                "state": {
                    "verbosity": app.state.verbosity,
                    "mode": app.state.mode,
                },
                "config": {
                    "ACCESS_LOG": app.config.ACCESS_LOG,
                    "NOISY_EXCEPTIONS": app.config.NOISY_EXCEPTIONS,
                },
                "shared_ctx": app.shared_ctx.__dict__,
            }
            for app in apps:
                kwargs["server_info"][app.name] = []
                for server_info in app.state.server_info:
                    server_info.settings = {
                        k: v
                        for k, v in server_info.settings.items()
                        if k not in ("main_start", "main_stop", "app", "ssl")
                    }
                    kwargs["server_info"][app.name].append(server_info)

            ssl = kwargs.get("ssl")

            if isinstance(ssl, SanicSSLContext):
                kwargs["ssl"] = ssl.sanic

            manager = WorkerManager(
                primary.state.workers,
                worker_serve,
                kwargs,
                cls._get_context(),
                (monitor_pub, monitor_sub),
                worker_state,
            )
            if cls.should_auto_reload():
                reload_dirs: set[Path] = primary.state.reload_dirs.union(
                    *(app.state.reload_dirs for app in apps)
                )
                reloader = Reloader(monitor_pub, 0, reload_dirs, app_loader)
                manager.manage("Reloader", reloader, {}, transient=False)

            inspector = None
            if primary.config.INSPECTOR:
                display, extra = primary.get_motd_data()
                packages = [
                    pkg.strip() for pkg in display["packages"].split(",")
                ]
                module = import_module("sanic")
                sanic_version = f"sanic=={module.__version__}"  # type: ignore
                app_info = {
                    **display,
                    "packages": [sanic_version, *packages],
                    "extra": extra,
                }
                inspector = primary.inspector_class(
                    monitor_pub,
                    app_info,
                    worker_state,
                    primary.config.INSPECTOR_HOST,
                    primary.config.INSPECTOR_PORT,
                    primary.config.INSPECTOR_API_KEY,
                    primary.config.INSPECTOR_TLS_KEY,
                    primary.config.INSPECTOR_TLS_CERT,
                )
                manager.manage("Inspector", inspector, {}, transient=False)

            primary._inspector = inspector
            primary._manager = manager

            ready = primary.listeners["main_process_ready"]
            trigger_events(ready, loop, primary)

            workers_started = True
            manager.run()
        except ServerKilled:
            exit_code = 1
        except BaseException:
            kwargs = primary_server_info.settings
            error_logger.error("Experienced exception while trying to serve")
            raise
        finally:
            logger.info("Server Stopped")
            for app in apps:
                app.state.server_info.clear()
                app.router.reset()
                app.signal_router.reset()

            for sock in socks:
                try:
                    sock.shutdown(SHUT_RDWR)
                except OSError:
                    ...
                sock.close()
            socks = []

            trigger_events(main_stop, loop, primary)

            loop.close()
            cls._cleanup_env_vars()
            cls._cleanup_apps()

            if workers_started:
                limit = 100
                while cls._get_process_states(worker_state):
                    sleep(0.1)
                    limit -= 1
                    if limit <= 0:
                        error_logger.warning(
                            "Worker shutdown timed out. "
                            "Some processes may still be running."
                        )
                        break
            sync_manager.shutdown()
            unix = kwargs.get("unix")
            if unix:
                remove_unix_socket(unix)
            logger.debug(get_goodbye())
        if exit_code:
            os._exit(exit_code)

    @staticmethod
    def _get_process_states(worker_state) -> list[str]:
        try:
            return [
                state
                for s in worker_state.values()
                if (
                    (state := s.get("state"))
                    and state
                    not in ("TERMINATED", "FAILED", "COMPLETED", "NONE")
                )
            ]
        except (BrokenPipeError, ConnectionResetError, EOFError):
            return []

    @classmethod
    def serve_single(cls, primary: Sanic | None = None) -> None:
        """Serve a single process of a Sanic application.

        Similar to `serve`, but only serves a single process. When used,
        certain features are disabled, such as `fast`, `workers`,
        `multiplexer`, `auto_reload`, and the Inspector. It is almost
        never needed to use this method directly. Instead, you should
        use the CLI:

        ```sh
        sanic app.sanic:app --single-process
        ```

        Or, if you need to do it programmatically, you should use the
        `single_process` argument of `run`:

        ```python
        app.run(single_process=True)
        ```

        Args:
            primary (Optional[Sanic], optional): The primary Sanic application
                to serve. Defaults to `None`.

        Raises:
            RuntimeError: Raised when no applications are found.
            RuntimeError: Raised when no server information is found for the
                primary application.
            RuntimeError: Raised when attempting to serve HTTP/3 as a
                secondary server.
            RuntimeError: Raised when attempting to use both `fast` and
                `workers`.
            ValueError: Raised when `PROXIES_COUNT` is negative.
        """
        os.environ["SANIC_MOTD_OUTPUT"] = "true"
        apps = list(cls._app_registry.values())

        if not primary:
            try:
                primary = apps[0]
            except IndexError:
                raise RuntimeError("Did not find any applications.")

        # This exists primarily for unit testing
        if not primary.state.server_info:  # no cov
            for app in apps:
                app.state.server_info.clear()
            return

        primary_server_info = primary.state.server_info[0]
        primary.before_server_start(partial(primary._start_servers, apps=apps))
        kwargs = {
            k: v
            for k, v in primary_server_info.settings.items()
            if k
            not in (
                "main_start",
                "main_stop",
                "app",
            )
        }
        kwargs["app_name"] = primary.name
        kwargs["app_loader"] = None
        sock = configure_socket(kwargs)

        kwargs["server_info"] = {}
        kwargs["server_info"][primary.name] = []
        for server_info in primary.state.server_info:
            server_info.settings = {
                k: v
                for k, v in server_info.settings.items()
                if k not in ("main_start", "main_stop", "app")
            }
            kwargs["server_info"][primary.name].append(server_info)

        try:
            worker_serve(monitor_publisher=None, **kwargs)
        except BaseException:
            error_logger.exception(
                "Experienced exception while trying to serve"
            )
            raise
        finally:
            logger.info("Server Stopped")
            for app in apps:
                app.state.server_info.clear()
                app.router.reset()
                app.signal_router.reset()

            if sock:
                sock.close()

            cls._cleanup_env_vars()
            cls._cleanup_apps()

    async def _start_servers(
        self,
        primary: Sanic,
        apps: list[Sanic],
    ) -> None:
        for app in apps:
            if (
                app.name is not primary.name
                and app.state.workers != primary.state.workers
                and app.state.server_info
            ):
                message = (
                    f"The primary application {repr(primary)} is running "
                    f"with {primary.state.workers} worker(s). All "
                    "application instances will run with the same number. "
                    f"You requested {repr(app)} to run with "
                    f"{app.state.workers} worker(s), which will be ignored "
                    "in favor of the primary application."
                )
                if is_atty():
                    message = "".join(
                        [
                            Colors.YELLOW,
                            message,
                            Colors.END,
                        ]
                    )
                error_logger.warning(message, exc_info=True)
            for server_info in app.state.server_info:
                if server_info.stage is not ServerStage.SERVING:
                    app.state.primary = False
                    handlers = [
                        *server_info.settings.pop("main_start", []),
                        *server_info.settings.pop("main_stop", []),
                    ]
                    if handlers:  # no cov
                        error_logger.warning(
                            f"Sanic found {len(handlers)} listener(s) on "
                            "secondary applications attached to the main "
                            "process. These will be ignored since main "
                            "process listeners can only be attached to your "
                            "primary application: "
                            f"{repr(primary)}"
                        )

                    if not server_info.settings["loop"]:
                        server_info.settings["loop"] = get_running_loop()

                    serve_args: dict[str, Any] = {
                        **server_info.settings,
                        "run_async": True,
                        "reuse_port": bool(primary.state.workers - 1),
                    }
                    if "app" not in serve_args:
                        serve_args["app"] = app
                    try:
                        server_info.server = await serve(**serve_args)
                    except OSError as e:  # no cov
                        first_message = (
                            "An OSError was detected on startup. "
                            "The encountered error was: "
                        )
                        second_message = str(e)
                        if is_atty():
                            message_parts = [
                                Colors.YELLOW,
                                first_message,
                                Colors.RED,
                                second_message,
                                Colors.END,
                            ]
                        else:
                            message_parts = [first_message, second_message]
                        message = "".join(message_parts)
                        error_logger.warning(message, exc_info=True)
                        continue
                    primary.add_task(
                        self._run_server(app, server_info), name="RunServer"
                    )

    async def _run_server(
        self,
        app: StartupMixin,
        server_info: ApplicationServerInfo,
    ) -> None:  # no cov
        try:
            # We should never get to this point without a server
            # This is primarily to keep mypy happy
            if not server_info.server:  # no cov
                raise RuntimeError("Could not locate AsyncioServer")
            if app.state.stage is ServerStage.STOPPED:
                server_info.stage = ServerStage.SERVING
                await server_info.server.startup()
                await server_info.server.before_start()
                await server_info.server.after_start()
            await server_info.server.serve_forever()
        except CancelledError:
            # We should never get to this point without a server
            # This is primarily to keep mypy happy
            if not server_info.server:  # no cov
                raise RuntimeError("Could not locate AsyncioServer")
            await server_info.server.before_stop()
            await server_info.server.close()
            await server_info.server.after_stop()
        finally:
            server_info.stage = ServerStage.STOPPED
            server_info.server = None

    @staticmethod
    def _cleanup_env_vars():
        variables = (
            "SANIC_RELOADER_PROCESS",
            "SANIC_IGNORE_PRODUCTION_WARNING",
            "SANIC_WORKER_NAME",
            "SANIC_MOTD_OUTPUT",
            "SANIC_WORKER_PROCESS",
            "SANIC_SERVER_RUNNING",
        )
        for var in variables:
            try:
                del os.environ[var]
            except KeyError:
                ...

    @classmethod
    def _cleanup_apps(cls):
        for app in cls._app_registry.values():
            app.state.server_info.clear()
            app.router.reset()
            app.signal_router.reset()
